"""
    Copyright 2018 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import pytest
import asyncpg
import datetime
import uuid
import time
import logging
import inspect
import types
import pkgutil

from inmanta import data, const
from inmanta.const import LogLevel
from asyncpg import PostgresSyntaxError


@pytest.mark.asyncio
async def test_postgres_client(postgresql_client):
    await postgresql_client.execute("CREATE TABLE test(id serial PRIMARY KEY, name VARCHAR (25) NOT NULL)")
    await postgresql_client.execute("INSERT INTO test VALUES(5, 'jef')")
    records = await postgresql_client.fetch("SELECT * FROM test")
    assert len(records) == 1
    first_record = records[0]
    assert first_record['id'] == 5
    assert first_record['name'] == "jef"
    await postgresql_client.execute("DELETE FROM test WHERE test.id = " + str(first_record['id']))
    records = await postgresql_client.fetch("SELECT * FROM test")
    assert len(records) == 0


@pytest.mark.asyncio
async def test_project(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    projects = await data.Project.get_list(name="test")
    assert len(projects) == 1
    assert projects[0].id == project.id

    other = await data.Project.get_by_id(project.id)
    assert project != other
    assert project.id == other.id


@pytest.mark.asyncio
async def test_project_unique(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    project = data.Project(name="test")
    with pytest.raises(asyncpg.UniqueViolationError):
        await project.insert()


def test_project_no_project_name(init_dataclasses_and_load_schema):
    with pytest.raises(AttributeError):
        data.Project()


@pytest.mark.asyncio
async def test_project_cascade_delete(init_dataclasses_and_load_schema):

    async def create_full_environment(project_name, environment_name):
        project = data.Project(name=project_name)
        await project.insert()

        env = data.Environment(name=environment_name, project=project.id, repo_url="", repo_branch="")
        await env.insert()

        agent_proc = data.AgentProcess(hostname="testhost",
                                       environment=env.id,
                                       first_seen=datetime.datetime.now(),
                                       last_seen=datetime.datetime.now(),
                                       sid=uuid.uuid4())
        await agent_proc.insert()

        agi1 = data.AgentInstance(process=agent_proc.sid, name="agi1", tid=env.id)
        await agi1.insert()
        agi2 = data.AgentInstance(process=agent_proc.sid, name="agi2", tid=env.id)
        await agi2.insert()

        agent = data.Agent(environment=env.id, name="agi1", last_failover=datetime.datetime.now(), paused=False,
                           primary=agi1.id)
        await agent.insert()

        version = int(time.time())
        cm = data.ConfigurationModel(version=version, environment=env.id)
        await cm.insert()

        resource_ids = []
        for i in range(5):
            path = "/etc/file" + str(i)
            key = "std::File[agent1,path=" + path + "]"
            res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                     attributes={"path": path})
            await res1.insert()
            resource_ids.append((res1.environment, res1.resource_version_id))

        code = data.Code(version=version, resource="std::File", environment=env.id)
        await code.insert()

        unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
        await unknown_parameter.insert()

        return project, env, agent_proc, [agi1, agi2], agent, resource_ids, code, unknown_parameter

    async def assert_project_exists(project, env, agent_proc, agent_instances, agent, resource_ids, code, unknown_parameter,
                                    exists):
        def func(x):
            if exists:
                return x is not None
            else:
                return x is None
        assert func(await data.Project.get_by_id(project.id))
        assert func(await data.Environment.get_by_id(env.id))
        assert func(await data.AgentProcess.get_one(sid=agent_proc.sid))
        assert func(await data.AgentInstance.get_by_id(agent_instances[0].id))
        assert func(await data.AgentInstance.get_by_id(agent_instances[1].id))
        assert func(await data.Agent.get_one(environment=agent.environment, name=agent.name))
        for (environment, resource_version_id) in resource_ids:
            assert func(await data.Resource.get_one(environment=environment, resource_version_id=resource_version_id))
        assert func(await data.Code.get_one(environment=code.environment, resource=code.resource, version=code.version))
        assert func(await data.UnknownParameter.get_by_id(unknown_parameter.id))

    # Setup two environments
    full_env_1 = await create_full_environment("proj1", "env1")
    full_env_2 = await create_full_environment("proj2", "env2")

    # Assert exists
    await assert_project_exists(*full_env_1, exists=True)
    await assert_project_exists(*full_env_2, exists=True)

    # Cascade delete project 1
    project_env1 = full_env_1[0]
    await project_env1.delete_cascade()

    # Assert outcome
    await assert_project_exists(*full_env_1, exists=False)
    await assert_project_exists(*full_env_2, exists=True)


@pytest.mark.asyncio
async def test_environment(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()
    assert env.project == project.id

    await project.delete_cascade()

    projects = await data.Project.get_list()
    envs = await data.Environment.get_list()
    assert len(projects) == 0
    assert len(envs) == 0


@pytest.mark.asyncio
async def test_environment_no_environment_name(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()
    with pytest.raises(AttributeError):
        data.Environment(project=project.id, repo_url="", repo_branch="")


@pytest.mark.asyncio
async def test_environment_no_project_id(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()
    with pytest.raises(AttributeError):
        data.Environment(name="dev", repo_url="", repo_branch="")


@pytest.mark.asyncio
async def test_environment_cascade_content_only(init_dataclasses_and_load_schema):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=uuid.uuid4())
    await agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.sid, name="agi1", tid=env.id)
    await agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.sid, name="agi2", tid=env.id)
    await agi2.insert()

    agent = data.Agent(environment=env.id, name="agi1", last_failover=datetime.datetime.now(), paused=False, primary=agi1.id)
    await agent.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(version=version, environment=env.id)
    await cm.insert()

    resource_ids = []
    for i in range(5):
        path = "/etc/file" + str(i)
        key = "std::File[agent1,path=" + path + "]"
        res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
        await res1.insert()
        resource_ids.append((res1.environment, res1.resource_version_id))

    code = data.Code(version=version, resource="std::File", environment=env.id)
    await code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    await unknown_parameter.insert()

    await env.set(data.AUTO_DEPLOY, True)

    assert (await data.Project.get_by_id(project.id)) is not None
    assert (await data.Environment.get_by_id(env.id)) is not None
    assert (await data.AgentProcess.get_one(sid=agent_proc.sid)) is not None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is not None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is not None
    assert (await data.Agent.get_one(environment=agent.environment, name=agent.name)) is not None
    for environment, resource_version_id in resource_ids:
        assert (await data.Resource.get_one(environment=environment, resource_version_id=resource_version_id)) is not None
    assert (await data.Code.get_one(environment=code.environment, resource=code.resource, version=code.version)) is not None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is not None
    assert (await env.get(data.AUTO_DEPLOY)) is True

    await env.delete_cascade(only_content=True)

    assert (await data.Project.get_by_id(project.id)) is not None
    assert (await data.Environment.get_by_id(env.id)) is not None
    assert (await data.AgentProcess.get_one(sid=agent_proc.sid)) is None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is None
    assert (await data.Agent.get_one(environment=agent.environment, name=agent.name)) is None
    for environment, resource_version_id in resource_ids:
        assert (await data.Resource.get_one(environment=environment, resource_version_id=resource_version_id)) is None
    assert (await data.Code.get_one(environment=code.environment, version=code.version)) is None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is None
    assert (await env.get(data.AUTO_DEPLOY)) is True


@pytest.mark.asyncio
async def test_environment_set_setting_parameter(init_dataclasses_and_load_schema):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    assert (await env.get(data.AUTO_DEPLOY)) is False
    await env.set(data.AUTO_DEPLOY, True)
    assert (await env.get(data.AUTO_DEPLOY)) is True
    await env.unset(data.AUTO_DEPLOY)
    assert (await env.get(data.AUTO_DEPLOY)) is False

    with pytest.raises(KeyError):
        await env.set("set_non_existing_parameter", 1)
    with pytest.raises(KeyError):
        await env.get("get_non_existing_parameter")
    with pytest.raises(AttributeError):
        await env.set(data.AUTO_DEPLOY, 5)


@pytest.mark.asyncio
async def test_environment_deprecated_setting(init_dataclasses_and_load_schema, caplog):
    project = data.Project(name="proj")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for (deprecated_option, new_option) in [(data.AUTOSTART_AGENT_INTERVAL, data.AUTOSTART_AGENT_DEPLOY_INTERVAL),
                                            (data.AUTOSTART_SPLAY, data.AUTOSTART_AGENT_DEPLOY_SPLAY_TIME)]:
        await env.set(deprecated_option, 22)
        caplog.clear()
        assert (await env.get(new_option)) == 22
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) in caplog.text

        await env.set(new_option, 23)
        caplog.clear()
        assert (await env.get(new_option)) == 23
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) not in caplog.text

        await env.unset(deprecated_option)
        caplog.clear()
        assert (await env.get(new_option)) == 23
        assert "Config option %s is deprecated. Use %s instead." % (deprecated_option, new_option) not in caplog.text


@pytest.mark.asyncio
async def test_agent_process(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1 = data.AgentInstance(process=agent_proc.sid, name="agi1", tid=env.id)
    await agi1.insert()
    agi2 = data.AgentInstance(process=agent_proc.sid, name="agi2", tid=env.id)
    await agi2.insert()

    agent_procs = await data.AgentProcess.get_by_env(env=env.id)
    assert len(agent_procs) == 1
    assert agent_procs[0].sid == agent_proc.sid

    assert (await data.AgentProcess.get_by_sid(sid)).sid == agent_proc.sid
    assert (await data.AgentProcess.get_by_sid(uuid.UUID(int=1))) is None

    live_procs = await data.AgentProcess.get_live()
    assert len(live_procs) == 1
    assert live_procs[0].sid == agent_proc.sid

    live_by_env_procs = await data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 1
    assert live_by_env_procs[0].sid == agent_proc.sid

    await agent_proc.update_fields(expired=datetime.datetime.now())

    live_procs = await data.AgentProcess.get_live()
    assert len(live_procs) == 0

    live_by_env_procs = await data.AgentProcess.get_live_by_env(env=env.id)
    assert len(live_by_env_procs) == 0

    await agent_proc.delete_cascade()

    assert (await data.AgentProcess.get_one(sid=agent_proc.sid)) is None
    assert (await data.AgentInstance.get_by_id(agi1.id)) is None
    assert (await data.AgentInstance.get_by_id(agi2.id)) is None


@pytest.mark.asyncio
async def test_agent_instance(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.sid, name=agi1_name, tid=env.id)
    await agi1.insert()
    agi2_name = "agi2"
    agi2 = data.AgentInstance(process=agent_proc.sid, name=agi2_name, tid=env.id)
    await agi2.insert()

    active_instances = await data.AgentInstance.active()
    assert len(active_instances) == 2
    assert agi1.id in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = await data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi1.id
    current_instances = await data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id

    await agi1.update_fields(expired=datetime.datetime.now())

    active_instances = await data.AgentInstance.active()
    assert len(active_instances) == 1
    assert agi1.id not in [x.id for x in active_instances]
    assert agi2.id in [x.id for x in active_instances]

    current_instances = await data.AgentInstance.active_for(env.id, agi1_name)
    assert len(current_instances) == 0
    current_instances = await data.AgentInstance.active_for(env.id, agi2_name)
    assert len(current_instances) == 1
    assert current_instances[0].id == agi2.id


@pytest.mark.asyncio
async def test_agent(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    sid = uuid.uuid4()
    agent_proc = data.AgentProcess(hostname="testhost",
                                   environment=env.id,
                                   first_seen=datetime.datetime.now(),
                                   last_seen=datetime.datetime.now(),
                                   sid=sid)
    await agent_proc.insert()

    agi1_name = "agi1"
    agi1 = data.AgentInstance(process=agent_proc.sid, name=agi1_name, tid=env.id)
    await agi1.insert()

    agent1 = data.Agent(environment=env.id, name="agi1_agent1", last_failover=datetime.datetime.now(), paused=False,
                        primary=agi1.id)
    await agent1.insert()
    agent2 = data.Agent(environment=env.id, name="agi1_agent2", paused=False)
    await agent2.insert()
    agent3 = data.Agent(environment=env.id, name="agi1_agent3", paused=True)
    await agent3.insert()

    agents = await data.Agent.get_list()
    assert len(agents) == 3
    for agent in agents:
        assert (agent.name, agent.environment) in [(a.name, a.environment) for a in [agent1, agent2, agent3]]

    for agent in [agent1, agent2, agent3]:
        retrieved_agent = await agent.get(agent.environment, agent.name)
        assert retrieved_agent is not None
        assert retrieved_agent.environment == agent.environment
        assert retrieved_agent.name == agent.name

    assert agent1.get_status() == "up"
    assert agent2.get_status() == "down"
    assert agent3.get_status() == "paused"

    for agent in [agent1, agent2, agent3]:
        assert agent.to_dict()["state"] == agent.get_status()

    await agent1.update_fields(paused=True)
    assert agent1.get_status() == "paused"

    await agent2.update_fields(primary=agi1.id)
    assert agent2.get_status() == "up"

    await agent3.update_fields(paused=False)
    assert agent3.get_status() == "down"

    primary_instance = await data.AgentInstance.get_by_id(agent1.primary)
    primary_process = await data.AgentProcess.get_one(sid=primary_instance.process)
    assert primary_process.sid == agent_proc.sid


@pytest.mark.asyncio
async def test_config_model(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(),
                                 total=1, version_info={})
    await cm.insert()

    # create resources
    key = "std::File[agent1,path=/etc/motd]"
    res1 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version, attributes={"path": "/etc/motd"})
    await res1.insert()

    agents = await data.ConfigurationModel.get_agents(env.id, version)
    assert len(agents) == 1
    assert "agent1" in agents


@pytest.mark.asyncio
async def test_model_list(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for version in range(1, 20):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        await cm.insert()

    versions = await data.ConfigurationModel.get_versions(env.id, 0, 1)
    assert len(versions) == 1
    assert versions[0].version == 19

    versions = await data.ConfigurationModel.get_versions(env.id, 1, 1)
    assert len(versions) == 1
    assert versions[0].version == 18

    versions = await data.ConfigurationModel.get_versions(env.id)
    assert len(versions) == 19
    assert versions[0].version == 19
    assert versions[-1].version == 1

    versions = await data.ConfigurationModel.get_versions(env.id, 10)
    assert len(versions) == 9
    assert versions[0].version == 9
    assert versions[-1].version == 1


@pytest.mark.asyncio
async def test_model_get_latest_version(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    cms = []
    for version in range(1, 5):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0,
                                     version_info={})
        await cm.insert()
        cms.append(cm)

    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version is None

    await cms[1].update_fields(released=True)
    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 2

    await cms[3].update_fields(released=True)
    latest_version = await data.ConfigurationModel.get_latest_version(env.id)
    assert latest_version.version == 4


@pytest.mark.asyncio
async def test_model_set_ready(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={})
    await cm.insert()

    assert cm.done == 0

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    await resource.insert()

    assert cm.done == 0
    await resource.update_fields(status=const.ResourceState.deployed)
    cm = await data.ConfigurationModel.get_one(version=version, environment=env.id)
    assert cm.done == 1


@pytest.mark.parametrize("resource_state, should_be_deployed", [
    (const.ResourceState.unavailable, True),
    (const.ResourceState.skipped, True),
    (const.ResourceState.deployed, True),
    (const.ResourceState.failed, True),
    (const.ResourceState.deploying, False),
    (const.ResourceState.available, False),
    (const.ResourceState.cancelled, True),
    (const.ResourceState.undefined, True),
    (const.ResourceState.skipped_for_undefined, True),
    (const.ResourceState.processing_events, False),
])
@pytest.mark.asyncio
async def test_model_mark_done_if_done(init_dataclasses_and_load_schema, resource_state, should_be_deployed):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={})
    await cm.insert()

    assert cm.done == 0

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    await resource.insert()

    assert not cm.deployed
    await data.ConfigurationModel.mark_done_if_done(env.id, cm.version)
    cm = await data.ConfigurationModel.get_one(version=version, environment=env.id)
    assert not cm.deployed
    assert cm.done == 0

    await resource.update_fields(status=resource_state)
    await data.ConfigurationModel.mark_done_if_done(env.id, cm.version)
    cm = await data.ConfigurationModel.get_one(version=version, environment=env.id)
    assert cm.deployed == should_be_deployed
    assert cm.done == (1 if should_be_deployed else 0)

    # Make sure that a done resource stays in done even when a repair is running
    await resource.update_fields(status=const.ResourceState.deploying)
    cm = await data.ConfigurationModel.get_one(version=version, environment=env.id)
    assert cm.deployed == should_be_deployed
    assert cm.done == (1 if should_be_deployed else 0)


@pytest.mark.asyncio
async def test_model_get_list(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env1 = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env1.insert()
    env2 = data.Environment(name="prod", project=project.id, repo_url="", repo_branch="")
    await env2.insert()

    for env in [env1, env2]:
        for i in range(2):
            cm = data.ConfigurationModel(environment=env.id, version=i, date=datetime.datetime.now(), total=0, version_info={})
            await cm.insert()

            for r in range(3):
                if r % 2 == 0:
                    res = data.Resource.new(environment=env.id, status=const.ResourceState.deployed,
                                            resource_version_id=f"std::File[agent1,path=/etc/file{r}],v={i}",
                                            attributes={"purge_on_delete": False}, last_deploy=datetime.datetime.now())
                else:
                    res = data.Resource.new(environment=env.id, status=const.ResourceState.deploying,
                                            resource_version_id=f"std::File[agent1,path=/etc/file{r}],v={i}",
                                            attributes={"purge_on_delete": False})
                await res.insert()

    for env in [env1, env2]:
        cms = await data.ConfigurationModel.get_list(environment=env.id)
        assert len(cms) == 2
        for c in cms:
            assert c.environment == env.id
            assert c.done == 2

    cms = await data.ConfigurationModel.get_list(environment=uuid.uuid4())
    assert not cms


@pytest.mark.asyncio
async def test_model_serialization(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    now = datetime.datetime.now()
    cm = data.ConfigurationModel(environment=env.id, version=version, date=now, total=1, version_info={})
    await cm.insert()

    assert cm.done == 0

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path}, status=const.ResourceState.deployed)
    await resource.insert()

    cm = await data.ConfigurationModel.get_one(environment=env.id, version=version)
    dct = cm.to_dict()
    assert dct['version'] == version
    assert dct['environment'] == env.id
    assert dct['date'] == now
    assert not dct['released']
    assert not dct['deployed']
    assert dct['result'] == const.VersionState.pending
    assert dct['version_info'] == {}
    assert dct['total'] == 1
    assert dct['done'] == 1
    assert dct['status'] == {str(uuid.uuid5(env.id, key)): {"id": key, "status": const.ResourceState.deployed.name}}


@pytest.mark.asyncio
async def test_model_delete_cascade(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    await cm.insert()

    path = "/etc/file"
    key = "std::File[agent1,path=" + path + "]"
    resource = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                                 attributes={"path": path})
    await resource.insert()

    code = data.Code(version=version, resource="std::File", environment=env.id)
    await code.insert()

    unknown_parameter = data.UnknownParameter(name="test", environment=env.id, version=version, source="")
    await unknown_parameter.insert()

    await cm.delete_cascade()

    assert (await data.ConfigurationModel.get_list()) == []
    assert (await data.Resource.get_one(environment=resource.environment,
                                        resource_version_id=resource.resource_version_id)) is None
    assert (await data.Code.get_one(environment=code.environment,
                                    resource=code.resource,
                                    version=code.version)) is None
    assert (await data.UnknownParameter.get_by_id(unknown_parameter.id)) is None


@pytest.mark.asyncio
async def test_model_get_version_nr_latest_version(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env_dev = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env_dev.insert()
    env_prod = data.Environment(name="prod", project=project.id, repo_url="", repo_branch="")
    await env_prod.insert()

    now = datetime.datetime.now()
    await data.ConfigurationModel(environment=env_dev.id, version=4, total=0, date=now, released=True).insert()
    await data.ConfigurationModel(environment=env_dev.id, version=7, total=0, date=now, released=True).insert()
    await data.ConfigurationModel(environment=env_dev.id, version=9, total=0, date=now, released=False).insert()

    await data.ConfigurationModel(environment=env_prod.id, version=15, total=0, date=now, released=False).insert()
    await data.ConfigurationModel(environment=env_prod.id, version=11, total=0, date=now, released=False).insert()

    assert await data.ConfigurationModel.get_version_nr_latest_version(env_dev.id) == 7
    assert await data.ConfigurationModel.get_version_nr_latest_version(env_prod.id) is None
    assert await data.ConfigurationModel.get_version_nr_latest_version(uuid.uuid4()) is None


@pytest.mark.parametrize("resource_state, version_state", [
    (const.ResourceState.deployed, const.VersionState.success),
    (const.ResourceState.failed, const.VersionState.failed),
    (const.ResourceState.undefined, const.VersionState.failed),
    (const.ResourceState.skipped_for_undefined, const.VersionState.failed),
    (const.ResourceState.cancelled, const.VersionState.failed),
    (const.ResourceState.skipped, const.VersionState.failed)
])
@pytest.mark.asyncio
async def test_mark_done(init_dataclasses_and_load_schema, resource_state, version_state):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=2, version_info={})
    await cm.insert()

    assert cm.done == 0

    path1 = "/etc/file1"
    key1 = "std::File[agent1,path=" + path1 + "]"
    resource1 = data.Resource.new(environment=env.id, resource_version_id=key1 + ",v=%d" % version,
                                  attributes={"path": path1}, status=resource_state)
    await resource1.insert()

    path2 = "/etc/file2"
    key2 = "std::File[agent1,path=" + path2 + "]"
    resource2 = data.Resource.new(environment=env.id, resource_version_id=key2 + ",v=%d" % version,
                                  attributes={"path": path2}, status=const.ResourceState.deployed)
    await resource2.insert()

    assert cm.result == const.VersionState.pending
    await cm.mark_done()
    assert cm.result == version_state


async def populate_model(env_id, version):

    def get_path(n):
        return "/tmp/%d" % n

    def get_id(n):
        return "std::File[agent1,path=/tmp/%d],v=%s" % (n, version)

    def get_resource(n, depends, status=const.ResourceState.available):
        requires = [get_id(z) for z in depends]
        return data.Resource.new(environment=env_id, resource_version_id=get_id(n),
                                 status=status,
                                 attributes={"path": get_path(n),
                                             "purge_on_delete": False,
                                             "purged": False,
                                             "requires": requires})

    res1 = get_resource(1, [])
    await res1.insert()

    res2 = get_resource(2, [1])
    await res2.insert()

    res3 = get_resource(3, [], const.ResourceState.undefined)
    await res3.insert()

    res4 = get_resource(4, [3])
    await res4.insert()

    res5 = get_resource(5, [4])
    await res5.insert()


@pytest.mark.asyncio
async def test_resource_purge_on_delete(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=2,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent2,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res12.insert()

    # model 2 (multiple undeployed versions)
    while version < 10:
        version += 1
        cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                      version_info={}, released=False, deployed=False)
        await cm2.insert()

        res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent5,path=/etc/motd],v=%s" % version,
                                  status=const.ResourceState.available,
                                  attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    await cm3.insert()

    to_purge = await data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.asyncio
async def test_issue_422(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    # model 2 (multiple undeployed versions)
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()

    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/motd],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res21.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=0, version_info={})
    await cm3.insert()

    to_purge = await data.Resource.get_deleted_resources(env.id, version, set())

    assert len(to_purge) == 1
    assert to_purge[0].model == 1
    assert to_purge[0].resource_id == "std::File[agent1,path=/etc/motd]"


@pytest.mark.asyncio
async def test_get_latest_resource(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    key = "std::File[agent1,path=/etc/motd]"
    assert (await data.Resource.get_latest_version(env.id, key)) is None

    version = 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()
    res11 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    await res11.insert()

    version = 2
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()
    res12 = data.Resource.new(environment=env.id, resource_version_id=key + ",v=%d" % version,
                              status=const.ResourceState.deployed,
                              attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res12.insert()

    res = await data.Resource.get_latest_version(env.id, key)
    assert res.model == 2


@pytest.mark.asyncio
async def test_get_resources(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    resource_ids = []
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id,
                                resource_version_id="std::File[agent1,path=/tmp/file%d],v=%d" % (i, version),
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids.append(res.resource_version_id)

    resources = await data.Resource.get_resources(env.id, resource_ids)
    assert len(resources) == len(resource_ids)
    assert sorted([x.resource_version_id for x in resources]) == sorted(resource_ids)

    resources = await data.Resource.get_resources(env.id, [resource_ids[0], "abcd"])
    assert len(resources) == 1

    resources = await data.Resource.get_resources(env.id, [])
    assert len(resources) == 0


@pytest.mark.asyncio
async def test_model_get_resources_for_version(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    resource_ids_version_one = []
    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    await cm.insert()
    for i in range(1, 11):
        res = data.Resource.new(environment=env.id,
                                resource_version_id="std::File[agent1,path=/tmp/file%d],v=%d" % (i, version),
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids_version_one.append(res.resource_version_id)

    resource_ids_version_two = []
    version += 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    await cm.insert()
    for i in range(11, 21):
        res = data.Resource.new(environment=env.id,
                                resource_version_id="std::File[agent2,path=/tmp/file%d],v=%d" % (i, version),
                                status=const.ResourceState.deployed,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        resource_ids_version_two.append(res.resource_version_id)

    for version in range(3, 5):
        cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                     version_info={}, released=True, deployed=True)
        await cm.insert()

    async def make_with_status(i, status):
        res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent3,path=/tmp/file%d],v=3" % i,
                                status=status,
                                attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
        await res.insert()
        return res.resource_version_id

    d = await make_with_status(1, const.ResourceState.deployed)
    s = await make_with_status(2, const.ResourceState.skipped)
    su = await make_with_status(3, const.ResourceState.skipped_for_undefined)
    u = await make_with_status(4, const.ResourceState.undefined)

    resources = await data.Resource.get_resources_for_version(env.id, 1)
    assert len(resources) == 10
    assert sorted(resource_ids_version_one) == sorted([x.resource_version_id for x in resources])
    resources = await data.Resource.get_resources_for_version(env.id, 2)
    assert len(resources) == 10
    assert sorted(resource_ids_version_two) == sorted([x.resource_version_id for x in resources])
    resources = await data.Resource.get_resources_for_version(env.id, 4)
    assert resources == []

    resources = await data.Resource.get_resources_for_version(env.id, 3)
    assert len(resources) == 4
    assert sorted([x.resource_version_id for x in resources]) == sorted([d, s, u, su])


@pytest.mark.asyncio
async def test_model_get_resources_for_version_optional_args(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=3, version_info={})
    await cm.insert()

    async def insert_resource(env_id, version, agent_name, path, status):
        resource_version_id = f"std::File[{agent_name},path={path}],v={version}"
        resource = data.Resource.new(environment=env_id,
                                     resource_version_id=resource_version_id,
                                     attributes={"path": path},
                                     status=status)
        await resource.insert()

    await insert_resource(env.id, version, "agent1", "path1", const.ResourceState.deployed)
    await insert_resource(env.id, version, "agent2", "path2", const.ResourceState.available)
    await insert_resource(env.id, version, "agent1", "path3", const.ResourceState.undefined)

    result = await data.Resource.get_resources_for_version(env.id, version)
    assert len(result) == 3
    assert sorted([r.agent for r in result]) == ["agent1", "agent1", "agent2"]
    for r in result:
        assert len(r.attributes) == 1

    result = await data.Resource.get_resources_for_version(env.id, version, agent="agent2")
    assert len(result) == 1
    assert result[0].agent == "agent2"

    result = await data.Resource.get_resources_for_version(env.id, version, no_obj=True)
    assert len(result) == 3
    assert sorted([r["agent"] for r in result]) == ["agent1", "agent1", "agent2"]
    for r in result:
        assert len(r["attributes"]) == 1


@pytest.mark.asyncio
async def test_escaped_resources(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    routes = {"8.0.0.0/8": "1.2.3.4", "0.0.0.0/0": "127.0.0.1"}
    res = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,name=router],v=%d" % version,
                            status=const.ResourceState.deployed,
                            attributes={"name": "router", "purge_on_delete": True, "purged": False, "routes": routes})
    await res.insert()
    resource_id = res.resource_version_id

    resources = await data.Resource.get_resources(env.id, [resource_id])
    assert len(resources) == 1

    assert resources[0].attributes["routes"] == routes


@pytest.mark.asyncio
async def test_resource_provides(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%d" % version,
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%d" % version,
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res1.provides.append(res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []

    await res1.insert()
    await res2.insert()

    res1 = await data.Resource.get(env.id, res1.resource_version_id)
    res2 = await data.Resource.get(env.id, res2.resource_version_id)

    assert len(res1.provides) == 1
    assert len(res2.provides) == 0
    assert res1.provides[0] == res2.resource_version_id
    assert res2.provides == []


@pytest.mark.asyncio
async def test_resource_hash(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for version in range(1, 4):
        cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                      version_info={}, released=True, deployed=True)
        await cm1.insert()

    res1 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=1",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res2 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=2",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": False})
    res3 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=3",
                             status=const.ResourceState.deployed,
                             attributes={"path": "/etc/motd", "purge_on_delete": True, "purged": True})
    await res1.insert()
    await res2.insert()
    await res3.insert()

    assert res1.attribute_hash is not None
    assert res1.attribute_hash == res2.attribute_hash
    assert res3.attribute_hash is not None
    assert res1.attribute_hash != res3.attribute_hash

    readres = await data.Resource.get_resources(env.id,
                                                [res1.resource_version_id, res2.resource_version_id, res3.resource_version_id])

    resource_map = {r.resource_version_id: r for r in readres}
    res1 = resource_map[res1.resource_version_id]
    res2 = resource_map[res2.resource_version_id]
    res3 = resource_map[res3.resource_version_id]

    assert res1.attribute_hash is not None
    assert res1.attribute_hash == res2.attribute_hash
    assert res3.attribute_hash is not None
    assert res1.attribute_hash != res3.attribute_hash


@pytest.mark.asyncio
async def test_resources_report(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # model 1
    version = 1
    cm1 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm1.insert()

    res11 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file1"})
    await res11.insert()
    res12 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 12, 30),
                              attributes={"path": "/etc/file2"})
    await res12.insert()

    # model 2
    version += 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=False, deployed=False)
    await cm2.insert()
    res21 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/file1"})
    await res21.insert()
    res22 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file3],v=%s" % version,
                              status=const.ResourceState.available,
                              attributes={"path": "/etc/file3"})
    await res22.insert()

    # model 3
    version += 1
    cm3 = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                  version_info={}, released=True, deployed=True)
    await cm3.insert()

    res31 = data.Resource.new(environment=env.id, resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                              status=const.ResourceState.deployed, last_deploy=datetime.datetime(2018, 7, 14, 14, 30),
                              attributes={"path": "/etc/file2"})
    await res31.insert()

    report = await data.Resource.get_resources_report(env.id)
    assert len(report) == 3
    report_as_map = {x["resource_id"]: x for x in report}
    for i in range(1, 4):
        assert f"std::File[agent1,path=/etc/file{i}]" in report_as_map

    assert report_as_map["std::File[agent1,path=/etc/file1]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["deployed_version"] == 1
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["latest_version"] == 2
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 12, 30)
    assert report_as_map["std::File[agent1,path=/etc/file1]"]["agent"] == "agent1"

    assert report_as_map["std::File[agent1,path=/etc/file2]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["deployed_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["latest_version"] == 3
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["last_deploy"] == datetime.datetime(2018, 7, 14, 14, 30)
    assert report_as_map["std::File[agent1,path=/etc/file2]"]["agent"] == "agent1"

    assert report_as_map["std::File[agent1,path=/etc/file3]"]["resource_type"] == "std::File"
    assert report_as_map["std::File[agent1,path=/etc/file3]"]["deployed_version"] is None
    assert report_as_map["std::File[agent1,path=/etc/file3]"]["latest_version"] == 2
    assert report_as_map["std::File[agent1,path=/etc/file3]"]["last_deploy"] is None
    assert report_as_map["std::File[agent1,path=/etc/file3]"]["agent"] == "agent1"


@pytest.mark.asyncio
async def test_resources_delete_cascade(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()
    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    await cm.insert()

    res1 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    await res1.insert()
    res2 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file2],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"path": "/etc/file1"})
    await res2.insert()
    action_id_resource_action_1 = uuid.uuid4()
    resource_action1 = data.ResourceAction(environment=env.id, resource_version_ids=[res1.resource_version_id],
                                           action_id=action_id_resource_action_1, action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    await resource_action1.insert()

    action_id_resource_action_2 = uuid.uuid4()
    resource_action2 = data.ResourceAction(environment=env.id, resource_version_ids=[res2.resource_version_id],
                                           action_id=action_id_resource_action_2, action=const.ResourceAction.deploy,
                                           started=datetime.datetime.now())
    await resource_action2.insert()

    await res1.delete_cascade()

    assert (await data.Resource.get_one(environment=res1.environment, resource_version_id=res1.resource_version_id)) is None
    assert (await data.ResourceAction.get_one(action_id=resource_action1.action_id)) is None
    resource = await data.Resource.get_one(environment=res2.environment, resource_version_id=res2.resource_version_id)
    assert resource is not None
    assert (resource.environment, resource.resource_version_id) == (res2.environment, res2.resource_version_id)
    resource_action = await data.ResourceAction.get_one(action_id=resource_action2.action_id)
    assert resource_action is not None
    assert resource_action.action_id == resource_action2.action_id
    resource_version_ids = await data.ResourceVersionId.get_list(environment=env.id)
    assert len(resource_version_ids) == 1
    assert resource_version_ids[0].action_id == action_id_resource_action_2


@pytest.mark.asyncio
async def test_resource_action(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    now = datetime.datetime.now()
    action_id = uuid.uuid4()
    resource_version_ids = ["std::File[agent1,path=/etc/file1],v=1", "std::File[agent1,path=/etc/file2],v=1"]
    resource_action = data.ResourceAction(environment=env.id, resource_version_ids=resource_version_ids, action_id=action_id,
                                          action=const.ResourceAction.deploy, started=now)
    await resource_action.insert()

    resource_action.add_changes({"rid": {"field1": {"old": "a", "new": "b"}, "field2": {}}})
    await resource_action.save()

    resource_action.add_changes({"rid": {"field2": {"old": "c", "new": "d"}, "field3": ['removed', 'installed']}})
    await resource_action.save()

    resource_action.add_logs([{}, {}])
    await resource_action.save()

    resource_action.add_logs([{}, {}])
    await resource_action.save()

    resource_action.set_field("status", const.ResourceState.failed)
    await resource_action.save()

    ra_via_get_by_id = await data.ResourceAction.get_one(action_id=resource_action.action_id)
    ra_list = await data.ResourceAction.get_list(action_id=resource_action.action_id)
    assert len(ra_list) == 1
    ra_via_get_list = ra_list[0]
    ra_via_get = await data.ResourceAction.get(action_id=resource_action.action_id)
    for ra in [ra_via_get_by_id, ra_via_get_list, ra_via_get]:
        assert ra.action_id == action_id
        assert ra.action == const.ResourceAction.deploy
        assert ra.started == now
        assert ra.finished is None

        assert len(ra.resource_version_ids) == 2
        assert sorted(ra.resource_version_ids) == sorted(resource_version_ids)

        assert len(ra.changes["rid"]) == 3
        assert ra.changes["rid"]["field1"]["old"] == "a"
        assert ra.changes["rid"]["field1"]["new"] == "b"
        assert ra.changes["rid"]["field2"]["old"] == "c"
        assert ra.changes["rid"]["field2"]["new"] == "d"
        assert ra.changes["rid"]["field3"] == ['removed', 'installed']
        assert ra.status == const.ResourceState.failed

        assert len(ra.messages) == 4
        for message in ra.messages:
            assert message == {}


@pytest.mark.asyncio
async def test_resource_action_get_logs(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    for i in range(1, 11):
        action_id = uuid.uuid4()
        resource_action = data.ResourceAction(environment=env.id,
                                              resource_version_ids=["std::File[agent1,path=/etc/motd],v=%1"],
                                              action_id=action_id,
                                              action=const.ResourceAction.deploy,
                                              started=datetime.datetime.now())
        await resource_action.insert()
        resource_action.add_logs([data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=i)])
        await resource_action.save()

    action_id = uuid.uuid4()

    resource_action = data.ResourceAction(environment=env.id,
                                          resource_version_ids=["std::File[agent1,path=/etc/motd],v=%1"],
                                          action_id=action_id,
                                          action=const.ResourceAction.dryrun,
                                          started=datetime.datetime.now())
    await resource_action.insert()
    times = datetime.datetime.now()
    resource_action.add_logs([data.LogLine.log(logging.WARNING, "warning version %(version)d", version=100, timestamp=times)])
    await resource_action.save()

    resource_actions = await data.ResourceAction.get_log(env.id,
                                                         "std::File[agent1,path=/etc/motd],v=%1")
    assert len(resource_actions) == 11
    for i in range(len(resource_actions)):
        action = resource_actions[i]
        if i == 0:
            assert action.action == const.ResourceAction.dryrun
        else:
            assert action.action == const.ResourceAction.deploy
    resource_actions = await data.ResourceAction.get_log(env.id,
                                                         "std::File[agent1,path=/etc/motd],v=%1",
                                                         const.ResourceAction.dryrun.name)
    assert len(resource_actions) == 1
    action = resource_actions[0]
    assert action.action == const.ResourceAction.dryrun
    assert action.messages[0]["level"] == LogLevel.WARNING.name
    assert action.messages[0]["timestamp"] == times
    resource_actions = await data.ResourceAction.get_log(env.id,
                                                         "std::File[agent1,path=/etc/motd],v=%1",
                                                         const.ResourceAction.deploy.name, limit=2)
    assert len(resource_actions) == 2
    for action in resource_actions:
        assert len(action.messages) == 1
        assert action.messages[0]["level"] == LogLevel.INFO.name


@pytest.mark.asyncio
async def test_data_document_recursion(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    now = datetime.datetime.now()
    ra = data.ResourceAction(environment=env.id,
                             resource_version_ids=["test"],
                             action_id=uuid.uuid4(),
                             action=const.ResourceAction.store,
                             started=now,
                             finished=now,
                             messages=[data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=2)])
    await ra.insert()


@pytest.mark.asyncio
async def test_code(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = int(time.time())
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={})
    await cm.insert()

    code1 = data.Code(environment=env.id, resource="std::File", version=version, source_refs={"ref": "ref"})
    await code1.insert()

    code2 = data.Code(environment=env.id, resource="std::Directory", version=version, source_refs={})
    await code2.insert()

    version2 = version + 1
    cm2 = data.ConfigurationModel(environment=env.id, version=version2, date=datetime.datetime.now(), total=1, version_info={})
    await cm2.insert()

    code3 = data.Code(environment=env.id, resource="std::Directory", version=version2, source_refs={})
    await code3.insert()

    def assert_match_code(code1, code2):
        assert code1 is not None
        assert code2 is not None
        assert code1.environment == code1.environment
        assert code1.resource == code2.resource
        assert code1.version == code2.version
        shared_keys_source_refs = [k for k in code1.source_refs
                                   if k in code2.source_refs and code1.source_refs[k] == code2.source_refs[k]]
        assert len(shared_keys_source_refs) == len(code1.source_refs.keys())

    code_file = await data.Code.get_version(env.id, version, "std::File")
    assert_match_code(code_file, code1)

    code_directory = await data.Code.get_version(env.id, version, "std::Directory")
    assert_match_code(code_directory, code2)

    code_test = await data.Code.get_version(env.id, version, "std::Test")
    assert code_test is None

    code_list = await data.Code.get_versions(env.id, version)
    ids_code_lost = [(c.environment, c.resource, c.version) for c in code_list]
    assert len(code_list) == 2
    assert (code1.environment, code1.resource, code1.version) in ids_code_lost
    assert (code2.environment, code2.resource, code2.version) in ids_code_lost
    code_list = await data.Code.get_versions(env.id, version + 1)
    assert len(code_list) == 1
    code = code_list[0]
    assert (code.environment, code.resource, code.version) == (code3.environment, code3.resource, code3.version)
    code_list = await data.Code.get_versions(env.id, version + 2)
    assert len(code_list) == 0


@pytest.mark.asyncio
async def test_parameter(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    time1 = datetime.datetime(2018, 7, 14, 12, 30)
    time2 = datetime.datetime(2018, 7, 16, 12, 30)
    time3 = datetime.datetime(2018, 7, 12, 12, 30)

    parameters = []
    for current_time in [time1, time2, time3]:
        t = current_time.strftime("%Y-%m-%dT%H:%M:%S.%f")
        parameter = data.Parameter(name="param_" + t, value="test_val_" + t, environment=env.id, source="test",
                                   updated=current_time)
        parameters.append(parameter)
        await parameter.insert()

    updated_before = await data.Parameter.get_updated_before(datetime.datetime(2018, 7, 12, 12, 30))
    assert len(updated_before) == 0
    updated_before = await data.Parameter.get_updated_before(datetime.datetime(2018, 7, 14, 12, 30))
    assert len(updated_before) == 1
    assert (updated_before[0].environment, updated_before[0].name) == (parameters[2].environment, parameters[2].name)
    updated_before = await data.Parameter.get_updated_before(datetime.datetime(2018, 7, 15, 12, 30))
    list_of_ids = [(x.environment, x.name) for x in updated_before]
    assert len(updated_before) == 2
    assert (parameters[0].environment, parameters[0].name) in list_of_ids
    assert (parameters[2].environment, parameters[2].name) in list_of_ids


@pytest.mark.asyncio
async def test_parameter_list_parameters(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    metadata_param1 = {"test1": "testval1", "test2": "testval2"}
    parameter1 = data.Parameter(name="param1", value="val", environment=env.id, source="test", metadata=metadata_param1)
    await parameter1.insert()

    metadata_param2 = {"test3": "testval3"}
    parameter2 = data.Parameter(name="param2", value="val", environment=env.id, source="test", metadata=metadata_param2)
    await parameter2.insert()

    results = await data.Parameter.list_parameters(env.id, **{"test1": "testval1"})
    assert len(results) == 1
    assert (results[0].environment, results[0].name) == (parameter1.environment, parameter1.name)

    results = await data.Parameter.list_parameters(env.id, **{"test1": "testval1", "test2": "testval2"})
    assert len(results) == 1
    assert (results[0].environment, results[0].name) == (parameter1.environment, parameter1.name)

    results = await data.Parameter.list_parameters(env.id, **{})
    assert len(results) == 2


@pytest.mark.asyncio
async def test_dryrun(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1, version_info={})
    await cm.insert()

    dryrun = await data.DryRun.create(env.id, version, 10, 5)

    resource_version_id = "std::File[agent1,path=/etc/motd],v=%s" % version
    dryrun_data = {"id": resource_version_id, "changes": {}}
    await data.DryRun.update_resource(dryrun.id, resource_version_id, dryrun_data)

    dryrun_retrieved = await data.DryRun.get_by_id(dryrun.id)
    assert dryrun_retrieved.todo == 4
    key = str(uuid.uuid5(dryrun.id, resource_version_id))
    assert dryrun_retrieved.resources[key]["changes"] == {}
    assert dryrun_retrieved.resources[key]["id"] == resource_version_id


@pytest.mark.asyncio
async def test_form(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    form = data.Form(environment=env.id, form_type="a type")
    await form.insert()
    other_form = data.Form(environment=env.id, form_type="other type")
    await other_form.insert()

    retrieved_form = await data.Form.get_form(env.id, "a type")
    assert retrieved_form is not None
    assert (retrieved_form.environment, retrieved_form.form_type) == (form.environment, form.form_type)

    non_existing_form = await data.Form.get_form(env.id, "non-existing-form")
    assert non_existing_form is None


@pytest.mark.asyncio
async def test_formrecord(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    form = data.Form(environment=env.id, form_type="a type")
    await form.insert()

    fields_for_record = {"field1": "val1", "field2": "val2"}
    changed = datetime.datetime.now()
    formrecord = data.FormRecord(environment=env.id, form=form.form_type, fields=fields_for_record, changed=changed)
    await formrecord.insert()

    results = await data.FormRecord.get_list()
    assert len(results) == 1
    result = results[0]
    assert result.id == formrecord.id
    assert len(result.fields) == 2
    for key, value in fields_for_record.items():
        assert result.fields[key] == value
    assert result.changed == formrecord.changed


@pytest.mark.asyncio
async def test_compile_get_reports(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    started1 = datetime.datetime(2018, 7, 14, 12, 30)
    started2 = datetime.datetime(2018, 7, 16, 12, 30)
    started3 = datetime.datetime(2018, 7, 12, 12, 30)

    compiles = []
    for started in [started1, started2, started3]:
        completed = started + datetime.timedelta(minutes=30)
        compile = data.Compile(environment=env.id, started=started, completed=completed)
        await compile.insert()
        compiles.append(compile)

    retrieved_compiles = await data.Compile.get_reports(env.id, None, None, None)
    assert len(retrieved_compiles) == 3
    assert retrieved_compiles[0]["id"] == compiles[1].id
    assert retrieved_compiles[1]["id"] == compiles[0].id
    assert retrieved_compiles[2]["id"] == compiles[2].id

    limit = 1
    retrieved_compiles = await data.Compile.get_reports(env.id, 1, None, None)
    assert len(retrieved_compiles) == 1
    assert retrieved_compiles[0]["id"] == compiles[1].id

    start_time = datetime.datetime(2018, 7, 13, 12, 30)
    retrieved_compiles = await data.Compile.get_reports(env.id, None, start_time, None)
    assert len(retrieved_compiles) == 2
    assert retrieved_compiles[0]["id"] == compiles[1].id
    assert retrieved_compiles[1]["id"] == compiles[0].id

    end_time = datetime.datetime(2018, 7, 15, 12, 30)
    retrieved_compiles = await data.Compile.get_reports(env.id, None, None, end_time)
    assert len(retrieved_compiles) == 2
    assert retrieved_compiles[0]["id"] == compiles[0].id
    assert retrieved_compiles[1]["id"] == compiles[2].id

    retrieved_compiles = await data.Compile.get_reports(env.id, limit, start_time, end_time)
    assert len(retrieved_compiles) == 1
    assert retrieved_compiles[0]["id"] == compiles[0].id


@pytest.mark.asyncio
async def test_compile_get_report(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # Compile 1
    started = datetime.datetime(2018, 7, 15, 12, 30)
    completed = datetime.datetime(2018, 7, 15, 13, 00)
    compile1 = data.Compile(environment=env.id, started=started, completed=completed)
    await compile1.insert()

    report_of_compile = await data.Compile.get_report(compile1.id)
    assert report_of_compile["reports"] == []

    report11 = data.Report(started=datetime.datetime.now(), completed=datetime.datetime.now(),
                           command="cmd", name="test", compile=compile1.id)
    await report11.insert()
    report12 = data.Report(started=datetime.datetime.now(), completed=datetime.datetime.now(),
                           command="cmd", name="test", compile=compile1.id)
    await report12.insert()

    # Compile 2
    compile2 = data.Compile(environment=env.id)
    await compile2.insert()
    report21 = data.Report(started=datetime.datetime.now(), completed=datetime.datetime.now(),
                           command="cmd", name="test", compile=compile2.id)
    await report21.insert()

    report_of_compile = await data.Compile.get_report(compile1.id)
    assert report_of_compile["id"] == compile1.id
    assert report_of_compile["started"] == started
    assert report_of_compile["completed"] == completed
    reports = report_of_compile["reports"]
    report_ids = [r["id"] for r in reports]
    assert len(reports) == 2
    assert report11.id in report_ids
    assert report12.id in report_ids

    report_of_compile = await data.Compile.get_report(compile2.id)
    reports = report_of_compile["reports"]
    assert len(reports) == 1


@pytest.mark.asyncio
async def test_match_tables_in_db_against_table_definitions_in_orm(postgres_db, database_name, postgresql_client,
                                                                   init_dataclasses_and_load_schema):
    table_names = await postgresql_client.fetch("SELECT table_name FROM information_schema.tables "
                                                "WHERE table_schema='public'")
    table_names_in_database = [x["table_name"] for x in table_names]
    table_names_in_classes_list = [x.__name__.lower() for x in data._classes]
    assert len(table_names_in_classes_list) == len(table_names_in_database)
    for item in table_names_in_classes_list:
        assert item in table_names_in_database


@pytest.mark.asyncio
async def test_dbschema_update_db_schema(postgresql_client, init_dataclasses_and_load_schema, get_columns_in_db_table):
    async def update_function1(connection):
        await connection.execute("CREATE TABLE public.tab(id integer primary key, val varchar NOT NULL);")

    async def update_function2(connection):
        await connection.execute("ALTER TABLE public.tab DROP COLUMN val;")

    current_db_version = await data.SchemaVersion.get_current_version()
    version_update1 = current_db_version + 1
    version_update2 = current_db_version + 2
    update_function_map = {version_update1: update_function1, version_update2: update_function2}

    db_schema = data.DBSchema()
    await db_schema._update_db_schema(update_function_map, postgresql_client)

    assert (await data.SchemaVersion.get_current_version()) == version_update2
    assert sorted(["id"]) == sorted(await get_columns_in_db_table("tab"))


@pytest.mark.asyncio
async def test_dbschema_update_db_schema_failure(postgresql_client, init_dataclasses_and_load_schema, get_columns_in_db_table):
    async def update_function(connection):
        # Syntax error should trigger database rollback
        await connection.execute("CREATE TABE public.tab(id integer primary key, val varchar NOT NULL);")

    current_db_version = await data.SchemaVersion.get_current_version()
    new_db_version = current_db_version + 1
    update_function_map = {new_db_version: update_function}

    db_schema = data.DBSchema()
    try:
        await db_schema._update_db_schema(update_function_map, postgresql_client)
    except PostgresSyntaxError:
        pass

    # Assert rollback
    assert (await data.SchemaVersion.get_current_version()) == current_db_version
    assert (await postgresql_client.fetchval("SELECT table_name FROM information_schema.tables "
                                             "WHERE table_schema='public' AND table_name='tab'")) is None

    async def update_function(connection):
        # Fix syntax issue
        await connection.execute("CREATE TABLE public.tab(id integer primary key, val varchar NOT NULL);")

    update_function_map[new_db_version] = update_function
    await db_schema._update_db_schema(update_function_map, postgresql_client)

    # Assert update
    assert (await data.SchemaVersion.get_current_version()) == new_db_version
    assert sorted(["id", "val"]) == sorted(await get_columns_in_db_table("tab"))


@pytest.mark.asyncio
async def test_dbschema_get_dct_with_update_functions():
    module_names = [modname for _, modname, ispkg in pkgutil.iter_modules(data.DBSchema.PACKAGE_WITH_UPDATE_FILES.__path__)
                    if not ispkg]
    all_versions = [int(mod_name[1:]) for mod_name in module_names]

    db_schema = data.DBSchema()
    update_function_map = await db_schema._get_dct_with_update_functions()
    assert sorted(all_versions) == sorted(update_function_map.keys())
    for version, update_function in update_function_map.items():
        assert version >= 0
        assert isinstance(update_function, types.FunctionType)
        assert update_function.__name__ == "update"
        assert inspect.getfullargspec(update_function)[0] == ["connection"]

    # Test behavior of "versions_higher_than" parameter
    lowest_version = min(update_function_map.keys())
    one_version_higher_than_lowest_version = lowest_version + 1

    restricted_function_map = await db_schema._get_dct_with_update_functions(one_version_higher_than_lowest_version)
    assert len(restricted_function_map) == len(update_function_map) - 1
    assert lowest_version not in restricted_function_map


@pytest.mark.asyncio
async def test_purgelog_test(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()

    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    # ResourceAction 1
    timestamp_ra1 = datetime.datetime.now() - datetime.timedelta(days=8)
    log_line_ra1 = data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=1)
    action_id = uuid.uuid4()
    ra1 = data.ResourceAction(environment=env.id, resource_version_ids=["id1"], action_id=action_id,
                              action=const.ResourceAction.store, started=timestamp_ra1, finished=datetime.datetime.now(),
                              messages=[log_line_ra1])
    await ra1.insert()

    # ResourceAction 2
    timestamp_ra2 = datetime.datetime.now() - datetime.timedelta(days=6)
    log_line_ra2 = data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=2)
    action_id = uuid.uuid4()
    ra2 = data.ResourceAction(environment=env.id, resource_version_ids=["id2"], action_id=action_id,
                              action=const.ResourceAction.store, started=timestamp_ra2, finished=datetime.datetime.now(),
                              messages=[log_line_ra2])
    await ra2.insert()

    assert len(await data.ResourceAction.get_list()) == 2
    await data.ResourceAction.purge_logs()
    assert len(await data.ResourceAction.get_list()) == 1
    remaining_resource_action = (await data.ResourceAction.get_list())[0]
    assert remaining_resource_action.action_id == ra2.action_id


@pytest.mark.asyncio
async def test_insert_many(init_dataclasses_and_load_schema, postgresql_client):
    project1 = data.Project(name="proj1")
    project2 = data.Project(name="proj2")
    projects = [project1, project2]
    await data.Project.insert_many(projects)

    result = await data.Project.get_list()
    project_names_in_result = [res.name for res in result]

    assert len(project_names_in_result) == 2
    assert sorted(["proj1", "proj2"]) == sorted(project_names_in_result)


@pytest.mark.asyncio
async def test_resources_json(init_dataclasses_and_load_schema):
    project = data.Project(name="test")
    await project.insert()
    env = data.Environment(name="dev", project=project.id, repo_url="", repo_branch="")
    await env.insert()

    version = 1
    cm = data.ConfigurationModel(environment=env.id, version=version, date=datetime.datetime.now(), total=1,
                                 version_info={}, released=True, deployed=True)
    await cm.insert()

    res1 = data.Resource.new(environment=env.id,
                             resource_version_id="std::File[agent1,path=/etc/file1],v=%s" % version,
                             status=const.ResourceState.deployed, last_deploy=datetime.datetime.now(),
                             attributes={"attr": [{"a": 1, "b": "c"}]})
    await res1.insert()

    res = await data.Resource.get_one(environment=res1.environment, resource_version_id=res1.resource_version_id)

    assert res1.attributes == res.attributes



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os
import time
import base64
import uuid
import argparse
from typing import Dict, List, Callable, Any, Union, Optional, Tuple, Set

from inmanta import protocol, const, loader
from inmanta.agent.handler import Commander
from inmanta.execute.util import Unknown, NoneValue
from inmanta.resources import resource, Resource, to_id, IgnoreResourceException
from inmanta.config import Option, is_uuid_opt, is_list, is_str
from inmanta.execute.proxy import DynamicProxy, UnknownException
from inmanta.ast import CompilerException, Locatable, OptionalValueException, type, Namespace
from inmanta.execute.runtime import Instance, ResultVariable
from inmanta.util import hash_file
import itertools
from inmanta.ast.entity import Entity
from inmanta.util import groupby
from inmanta.ast.attribute import RelationAttribute
import inmanta.model as model


LOGGER = logging.getLogger(__name__)

unknown_parameters = []

cfg_env = Option("config", "environment", None, "The environment this model is associated with", is_uuid_opt)
cfg_export = Option("config", "export", "", "The list of exporters to use", is_list)
cfg_unknown_handler = Option("unknown_handler", "default", "prune-agent", "default method to handle unknown values ", is_str)


ModelDict = Dict[str, List[type.Type]]
ResourceDict = Dict[str, Resource]
ProxiedType = Dict[str, List[Union[str, tuple, int, float, bool, "DynamicProxy"]]]


class DependencyCycleException(Exception):

    def __init__(self, start: Resource) -> None:
        super().__init__()
        self.start = start
        self.cycle = [start]
        self.running = True

    def add_to_cycle(self, node: Resource) -> None:
        if node == self.start:
            self.running = False
        elif self.running:
            self.cycle.append(node)

    def __str__(self) -> str:
        return "Cycle in dependencies: %s" % self.cycle


def upload_code(conn: protocol.Client, tid: uuid.UUID, version: int, code_manager: loader.CodeManager) -> None:
    res = conn.stat_files(list(code_manager.get_file_hashes()))
    if res is None or res.code != 200:
        raise Exception("Unable to upload handler plugin code to the server (msg: %s)" % res.result)

    for file in res.result["files"]:
        content = code_manager.get_file_content(file)
        res = conn.upload_file(id=file, content=base64.b64encode(content.encode()).decode("ascii"))
        if res is None or res.code != 200:
            raise Exception("Unable to upload handler plugin code to the server (msg: %s)" % res.result)

    source_map = {
        resource_name: {
            source.hash: (source.path, source.module_name, source.requires) for source in sources
        } for resource_name, sources in code_manager.get_types()
    }

    res = conn.upload_code_batched(tid=tid, id=version, resources=source_map)
    if res is None or res.code != 200:
        raise Exception("Unable to upload handler plugin code to the server (msg: %s)" % res.result)


class Exporter(object):
    """
        This class handles exporting the compiled configuration model
    """
    # instance vars
    types: Optional[Dict[str, Entity]]
    scopes: Optional[Namespace]

    # class vars
    __export_functions: Dict[str, Tuple[List[str], Callable[["Exporter", ProxiedType], None]]] = {}
    __dep_manager: List[Callable[[ModelDict, ResourceDict], None]] = []

    @classmethod
    def add(cls, name: str, types: List[str], function: Callable[["Exporter", ProxiedType], None]) -> None:
        """
            Add a new export function
        """
        cls.__export_functions[name] = (types, function)

    @classmethod
    def add_dependency_manager(cls, function: Callable[[ModelDict, ResourceDict], None]) -> None:
        """
            Register a new dependency manager
        """
        cls.__dep_manager.append(function)

    def __init__(self, options: argparse.Namespace=None) -> None:
        self.options = options

        self._resources: ResourceDict = {}
        self._resource_to_host = {}
        self._resource_state = {}
        self._unknown_objects = set()
        self._version = 0
        self._scope = None

        self._file_store = {}

    def _get_instance_proxies_of_types(self, types: List[str]) -> Dict[str, ProxiedType]:
        """ Returns a dict of instances for the given types
        """
        proxies = {}
        for t in types:
            if self.types is not None and t in self.types:
                proxies[t] = [DynamicProxy.return_value(i) for i in self.types[t].get_all_instances()]
            else:
                proxies[t] = []

        return proxies

    def _load_resources(self, types: Dict[str, Entity]) -> None:
        """
            Load all registered resources
        """
        resource.validate()
        entities = resource.get_entity_resources()
        resource_mapping = {}
        ignored_set = set()

        for entity in entities:
            if entity not in types:
                continue
            instances = types[entity].get_all_instances()
            if len(instances) > 0:
                for instance in instances:
                    try:
                        res = Resource.create_from_model(self, entity, DynamicProxy.return_value(instance))
                        resource_mapping[instance] = res
                        self.add_resource(res)
                    except UnknownException:
                        ignored_set.add(instance)
                        # We get this exception when the attribute that is used to create the object id contains an unknown.
                        # We can safely ignore this resource == prune it
                        LOGGER.debug("Skipped resource of type %s because its id contains an unknown (location: %s)",
                                     entity, instance.location)

                    except IgnoreResourceException:
                        ignored_set.add(instance)
                        LOGGER.info("Ignoring resource of type %s because it requested to ignore it. (location: %s)",
                                    entity, instance.location)

        Resource.convert_requires(resource_mapping, ignored_set)

    def _run_export_plugins(self) -> None:
        """
            Run any additional export plug-ins
        """
        export = []
        for pl in cfg_export.get():
            export.append(pl.strip())

        for name in export:
            if name.strip() == '':
                continue

            if name not in self.__class__.__export_functions:
                raise Exception("Export function %s does not exist." % name)

            types, function = Exporter.__export_functions[name]

            if len(types) > 0:
                function(self, types=self._get_instance_proxies_of_types(types))
            else:
                function(self)

    def _call_dep_manager(self, types: ModelDict) -> None:
        """
            Call all dep managers and let them add dependencies
        """
        for fnc in self.__class__.__dep_manager:
            try:
                fnc(types, self._resources)
            except UnknownException:
                LOGGER.debug("Dependency manager %s caused an unknown exception", fnc)

        # TODO: check for cycles

    def _validate_graph(self) -> None:
        """
            Validate the graph and if requested by the user, dump it
        """
        done: Set[Resource] = set()

        def find_cycle(current: Resource, working: Set[Resource]) -> None:
            if current in done:
                return
            if current in working:
                raise DependencyCycleException(current)
            working.add(current)
            for dep in current.requires:
                try:
                    find_cycle(dep, working)
                except DependencyCycleException as e:
                    e.add_to_cycle(current)
                    raise e
            done.add(current)
            working.remove(current)

        for res in self._resources.values():
            if res not in done:
                find_cycle(res, set())

        if self.options and self.options.depgraph:
            dot = "digraph G {\n"
            for res in self._resources.values():
                res_id = str(res.id)
                dot += '\t"%s";\n' % res_id

                for req in res.requires:
                    dot += '\t"%s" -> "%s";\n' % (res_id, str(req))

            dot += "}\n"

            with open("dependencies.dot", "wb+") as fd:
                fd.write(dot.encode())

    def run(
        self,
        types: Optional[Dict[str, type.Type]],
        scopes: Optional[Namespace],
        metadata: Dict[str, str]={},
        no_commit: bool=False,
        include_status: bool=False,
        model_export: bool=False
    ) -> None:
        """
        Run the export functions
        """
        self.types = types
        self.scopes = scopes
        self._version = int(time.time())

        # first run other export plugins
        self._run_export_plugins()

        if types is not None:
            # then process the configuration model to submit it to the mgmt server
            self._load_resources(types)

            # call dependency managers
            self._call_dep_manager(types)
            metadata[const.META_DATA_COMPILE_STATE] = const.Compilestate.success.name
        else:
            metadata[const.META_DATA_COMPILE_STATE] = const.Compilestate.failed.name

        # validate the dependency graph
        self._validate_graph()

        resources = self.resources_to_list()

        if len(self._resources) == 0:
            LOGGER.warning("Empty deployment model.")

        model = {}

        if self.options and self.options.json:
            with open(self.options.json, "wb+") as fd:
                fd.write(protocol.json_encode(resources).encode("utf-8"))
            if len(self._resources) > 0 or len(unknown_parameters) > 0:
                model = ModelExporter(types).export_all()
                with open(self.options.json + ".types", "wb+") as fd:
                    fd.write(protocol.json_encode(model).encode("utf-8"))
        elif len(self._resources) > 0 or len(unknown_parameters) > 0 and not no_commit:
            model = None
            if types is not None and model_export:
                model = ModelExporter(types).export_all()

            self.commit_resources(self._version, resources, metadata, model)
            LOGGER.info("Committed resources with version %d" % self._version)

        if include_status:
            return self._version, self._resources, self._resource_state, model
        return self._version, self._resources

    def add_resource(self, resource: Resource) -> None:
        """
            Add a new resource to the list of exported resources. When
            commit_resources is called, the entire list of resources is send
            to the the server.

            A resource is a map of attributes. This method validates the id
            of the resource and will add a version (if it is not set already)
        """
        if resource.version > 0:
            raise Exception("Versions should not be added to resources during model compilation.")

        resource.set_version(self._version)

        if resource.id in self._resources:
            raise CompilerException("Resource %s exists more than once in the configuration model" % resource.id)

        is_undefined = False
        for unknown in resource.unknowns:
            is_undefined = True
            value = getattr(resource, unknown)
            if value.source is not None and hasattr(value.source, "_type"):
                self._unknown_objects.add(to_id(value.source))

        if is_undefined:
            self._resource_state[resource.id.resource_str()] = const.ResourceState.undefined
        else:
            self._resource_state[resource.id.resource_str()] = const.ResourceState.available

        self._resources[resource.id] = resource
        self._resource_to_host[resource.id] = resource.id.agent_name

    def resources_to_list(self) -> List[Dict[str, Any]]:
        """ Convert the resource list to a json representation
        """
        resources = []

        for res in self._resources.values():
            resources.append(res.serialize())

        return resources

    def deploy_code(self, conn: protocol.Client, tid: uuid, version: int=None) -> None:
        """ Deploy code to the server
        """
        if version is None:
            version = int(time.time())

        code_manager = loader.CodeManager()
        LOGGER.info("Sending resources and handler source to server")

        # Load both resource definition and handlers
        for type_name, resource_definition in resource.get_resources():
            code_manager.register_code(type_name, resource_definition)

        for type_name, handler_definition in Commander.get_providers():
            code_manager.register_code(type_name, handler_definition)

        LOGGER.info("Uploading source files")

        upload_code(conn, tid, version, code_manager)

    def commit_resources(
        self, version: int, resources: List[Dict[str, str]], metadata: Dict[str, str], model: Dict
    ) -> None:
        """
            Commit the entire list of resource to the configurations server.
        """
        tid = cfg_env.get()
        if tid is None:
            LOGGER.error("The environment for this model should be set!")
            return

        conn = protocol.SyncClient("compiler")
        self.deploy_code(conn, tid, version)

        LOGGER.info("Uploading %d files" % len(self._file_store))

        # collect all hashes and send them at once to the server to check
        # if they are already uploaded
        hashes = list(self._file_store.keys())

        res = conn.stat_files(files=hashes)

        if res.code != 200:
            raise Exception("Unable to check status of files at server")

        to_upload = res.result["files"]

        LOGGER.info("Only %d files are new and need to be uploaded" % len(to_upload))
        for hash_id in to_upload:
            content = self._file_store[hash_id]

            res = conn.upload_file(id=hash_id, content=base64.b64encode(content).decode("ascii"))

            if res.code != 200:
                LOGGER.error("Unable to upload file with hash %s" % hash_id)
            else:
                LOGGER.debug("Uploaded file with hash %s" % hash_id)

        # Collecting version information
        version_info = {const.EXPORT_META_DATA: metadata,
                        "model": model}

        # TODO: start transaction
        LOGGER.info("Sending resource updates to server")
        for res in resources:
            LOGGER.debug("  %s", res["id"])

        res = conn.put_version(tid=tid, version=version, resources=resources, unknowns=unknown_parameters,
                               resource_state=self._resource_state, version_info=version_info)

        if res.code != 200:
            LOGGER.error("Failed to commit resource updates (%s)", res.result["message"])
            raise Exception("Failed to commit resource updates (%s)" % res.result["message"])

    def upload_file(self, content: Union[str, bytes] = None) -> str:
        """
            Upload a file to the configuration server. This operation is not
            executed in the transaction.
        """
        if not isinstance(content, bytes):
            content = content.encode('utf-8')

        hash_id = hash_file(content)
        self._file_store[hash_id] = content

        return hash_id


class dependency_manager(object):  # noqa: N801
    """
    Register a function that manages dependencies in the configuration model that will be deployed.
    """

    def __init__(self, function: Callable[[ModelDict, ResourceDict], None]) -> None:
        Exporter.add_dependency_manager(function)


class code_manager(object):  # noqa: N801
    """ Register a function that will be invoked after all resource and handler code is collected. A code manager can add
    or modify code before it is uploaded to the server.
    """
    def __init__(self, function: Callable[[], None]) -> None:
        pass


class export(object):  # noqa: N801
    """
        A decorator that registers an export function
    """

    def __init__(self, name: str, *args: str) -> None:
        self.name = name
        self.types = args

    def __call__(self, function: Callable[["Exporter", ProxiedType], None]) -> Callable[["Exporter", ProxiedType], None]:
        """
            The wrapping
        """
        Exporter.add(self.name, self.types, function)
        return function


@export("dump", "std::File", "std::Service", "std::Package")
def export_dumpfiles(exporter: Exporter, types: ProxiedType) -> None:
    prefix = "dump"

    if not os.path.exists(prefix):
        os.mkdir(prefix)

    for file in types["std::File"]:
        path = os.path.join(prefix, file.host.name + file.path.replace("/", "+"))
        with open(path, "w+") as fd:
            if isinstance(file.content, Unknown):
                fd.write("UNKNOWN -> error")
            else:
                fd.write(file.content)

    path = os.path.join(prefix, "services")
    with open(path, "w+") as fd:
        for svc in types["std::Service"]:
            fd.write("%s -> %s\n" % (svc.host.name, svc.name))

    path = os.path.join(prefix, "packages")
    with open(path, "w+") as fd:
        for pkg in types["std::Package"]:
            fd.write("%s -> %s\n" % (pkg.host.name, pkg.name))


def location(obj: Locatable) -> model.Location:
    loc = obj.get_location()
    return model.Location(loc.file, loc.lnr)


def relation_name(type: Entity, rel: RelationAttribute) -> str:
    if rel is None:
        return ""
    return type.get_full_name() + "." + rel.name


class ModelExporter(object):

    def __init__(self, types: ModelDict):
        self.root_type = types["std::Entity"]
        self.types = types

    def export_types(self) -> Dict[str, Any]:
        """
            Run after export_model!!
        """
        def convert_comment(value):
            if value is None:
                return ''
            else:
                return str(value)

        def convert_value_for_type(value):
            if isinstance(value, Unknown):
                raise Exception("annotations should not be unknown")
            if isinstance(value, Instance):
                return model.ReferenceValue(self.entity_ref[value])
            else:
                return model.DirectValue(value)

        def convert_attribute(attr):
            return model.Attribute(attr.type.type_string(), attr.is_optional(), attr.is_multi(),
                                   convert_comment(attr.comment), location(attr))

        def convert_relation(relation: RelationAttribute):

            return model.Relation(relation.type.get_full_name(),
                                  (relation.low, relation.high),
                                  relation_name(relation.type, relation.end),
                                  convert_comment(relation.comment), location(relation),
                                  [convert_value_for_type(x.get_value()) for x in relation.source_annotations],
                                  [convert_value_for_type(x.get_value()) for x in relation.target_annotations])

        def convert_type(mytype):
            return model.Entity([x.get_full_name() for x in mytype.parent_entities],
                                {n: convert_attribute(attr) for n, attr in mytype.get_attributes().items()
                                 if not isinstance(attr, RelationAttribute)},
                                {n: convert_relation(attr) for n, attr in mytype.get_attributes().items()
                                 if isinstance(attr, RelationAttribute)},
                                location(mytype))
        return {k: convert_type(v).to_dict() for k, v in self.types.items() if isinstance(v, Entity)}

    def export_model(self) -> Dict[str, Any]:
        entities = self.root_type.get_all_instances()

        entities_per_type = {t: [e for e in g] for t, g in groupby(entities, lambda x: x.type.get_full_name())}

        entity_ref = {e: t + "_" + str(i) for t, es in entities_per_type.items() for e, i in zip(es, itertools.count(1))}
        self.entity_ref = entity_ref

        def convert(value):
            if isinstance(value, Unknown):
                return "_UNKNOWN_"
            if isinstance(value, Instance):
                return entity_ref[value]
            return value

        def convert_relation(value: ResultVariable):
            if value.is_ready() and value.value is not None:
                rawvalue = value.get_value()
            else:
                # no value present
                return {"values": []}
            if not isinstance(rawvalue, list):
                rawvalue = [rawvalue]
            return {"values": [convert(v) for v in rawvalue]}

        def convert_attribute(value: ResultVariable):
            try:
                rawvalue = value.get_value()
            except OptionalValueException:
                return {"nones": [0]}
            if isinstance(rawvalue, Unknown):
                return {"unknowns": [0]}
            if isinstance(rawvalue, NoneValue):
                return {"nones": [0]}
            if not isinstance(rawvalue, list):
                rawvalue = [rawvalue]
                return {"values": rawvalue}
            else:
                unknowns = []
                offset = 0
                for i in range(0, len(rawvalue)):
                    value = rawvalue[i - offset]
                    if isinstance(value, Unknown):
                        unknowns.append(i)
                        del rawvalue[i - offset]
                        offset += 1
                if len(unknowns) > 0:
                    return {"values": rawvalue, "unknowns": unknowns}
                else:
                    return {"values": rawvalue}

        def convert_entity(original):
            attributes = {}
            relations = {}
            map = {"type": original.type.get_full_name(), "relations": relations, "attributes": attributes}
            for name, value in original.slots.items():
                if name == "self":
                    pass
                elif isinstance(value.type, Entity):
                    relations[name] = convert_relation(value)
                else:
                    attributes[name] = convert_attribute(value)
            return map

        maps = {entity_ref[k]: convert_entity(k) for k in entities}

        return maps

    def export_all(self) -> Dict[str, Any]:
        return {
            "instances": self.export_model(),
            "types": self.export_types()
        }



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import functools
import hashlib
import inspect
import itertools
import logging
import socket
import warnings
import uuid
import datetime
import enum
from asyncio import ensure_future, CancelledError
from logging import Logger

import pkg_resources
from pkg_resources import DistributionNotFound
from tornado.ioloop import IOLoop
from typing import Callable, Dict, Union, Tuple, List, Coroutine
from tornado import gen

from inmanta.types import JsonType

LOGGER = logging.getLogger(__name__)
SALT_SIZE = 16
HASH_ROUNDS = 100000


def memoize(obj):
    cache = obj.cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        if args not in cache:
            cache[args] = obj(*args, **kwargs)
        return cache[args]

    return memoizer


def get_compiler_version() -> str:
    try:
        return pkg_resources.get_distribution("inmanta").version
    except DistributionNotFound:
        LOGGER.error(
            "Could not find version number for the inmanta compiler."
            "Is inmanta installed? Use stuptools install or setuptools dev to install."
        )
        return None


def groupby(mylist, f):
    return itertools.groupby(sorted(mylist, key=f), f)


def hash_file(content: str) -> str:
    """
        Create a hash from the given content
    """
    sha1sum = hashlib.new("sha1")
    sha1sum.update(content)

    return sha1sum.hexdigest()


def is_call_ok(result: Union[int, Tuple[int, JsonType]]) -> bool:
    if isinstance(result, tuple):
        if len(result) == 2:
            code, reply = result
        else:
            raise Exception("Handlers for method call can only return a status code and a reply")

    else:
        code = result

    return code == 200


def ensure_future_and_handle_exception(logger: Logger, msg: str, action: Union[Coroutine]) -> None:
    """ Fire off a coroutine from the ioloop thread and log exceptions to the logger with the message """
    future = ensure_future(action)

    def handler(future):
        try:
            exc = future.exception()
            if exc is not None:
                logger.exception(msg, exc_info=exc)
        except CancelledError:
            pass

    future.add_done_callback(handler)


class Scheduler(object):
    """
        An event scheduler class
    """
    def __init__(self, name: str) -> None:
        self.name = name
        self._scheduled: Dict[Callable, object] = {}

    def add_action(self, action: Union[Callable, Coroutine], interval: float, initial_delay: float = None) -> None:
        """
            Add a new action

            :param action: A function to call periodically
            :param interval: The interval between execution of actions
            :param initial_delay: Delay to the first execution, defaults to interval
        """
        assert inspect.iscoroutinefunction(action) or gen.is_coroutine_function(action)

        if initial_delay is None:
            initial_delay = interval

        LOGGER.debug("Scheduling action %s every %d seconds with initial delay %d", action, interval, initial_delay)

        def action_function():
            LOGGER.info("Calling %s" % action)
            if action in self._scheduled:
                try:
                    ensure_future_and_handle_exception(
                        LOGGER,
                        "Uncaught exception while executing scheduled action",
                        action()
                    )
                except Exception:
                    LOGGER.exception("Uncaught exception while executing scheduled action")
                finally:
                    # next iteration
                    ihandle = IOLoop.current().call_later(interval, action_function)
                    self._scheduled[action] = ihandle

        handle = IOLoop.current().call_later(initial_delay, action_function)
        self._scheduled[action] = handle

    def remove(self, action: Callable) -> None:
        """
            Remove a scheduled action
        """
        if action in self._scheduled:
            IOLoop.current().remove_timeout(self._scheduled[action])
            del self._scheduled[action]

    def stop(self) -> None:
        """
            Stop the scheduler
        """
        try:
            # remove can still run during stop. That is why we loop until we get a keyerror == the dict is empty
            while True:
                action, handle = self._scheduled.popitem()
                IOLoop.current().remove_timeout(handle)
        except KeyError:
            pass

    def __del__(self) -> None:
        if len(self._scheduled) > 0:
            warnings.warn("Deleting scheduler '%s' that has not been stopped properly." % self.name)


def get_free_tcp_port() -> str:
    """
        Semi safe method for getting a random port. This may contain a race condition.
    """
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    _addr, port = tcp.getsockname()
    tcp.close()
    return str(port)


def custom_json_encoder(o: object) -> Union[Dict, str, List]:
    """
        A custom json encoder that knows how to encode other types commonly used by Inmanta from standard python libraries
    """
    if isinstance(o, uuid.UUID):
        return str(o)

    if isinstance(o, datetime.datetime):
        return o.isoformat(timespec='microseconds')

    if hasattr(o, "to_dict"):
        return o.to_dict()

    if isinstance(o, enum.Enum):
        return o.name

    if isinstance(o, Exception):
        # Logs can push exceptions through RPC. Return a string representation.
        return str(o)

    LOGGER.error("Unable to serialize %s", o)
    raise TypeError(repr(o) + " is not JSON serializable")



from inmanta.const import (
    TRANSIENT_STATES,
    UNDEPLOYABLE_STATES,
    NOT_DONE_STATES,
    VALID_STATES_ON_STATE_UPDATE,
    DONE_STATES,
    INITIAL_STATES,
)


def test_action_set_consistency():

    undep = set(UNDEPLOYABLE_STATES)
    transient = set(TRANSIENT_STATES)
    not_done = set(NOT_DONE_STATES)
    done = set(DONE_STATES)

    initial = set(INITIAL_STATES)
    on_deploy = set(VALID_STATES_ON_STATE_UPDATE)

    # transient is not done
    assert transient <= not_done
    # undeployed is done
    assert undep <= done

    # done + not_done == all == initial states + states one can transition to
    assert done | not_done == on_deploy | initial



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

from inmanta.ast import Locatable, RuntimeException, TypingException
from inmanta.ast.type import TypedList, NullableType
from inmanta.execute.runtime import ResultVariable, ListVariable, OptionVariable, AttributeVariable, QueueScheduler
from inmanta.execute.util import Unknown
from typing import List, Tuple, Optional

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
    from inmanta.ast.type import Type  # noqa: F401
    from inmanta.execute.runtime import Instance  # noqa: F401
    from inmanta.ast.entity import Entity  # noqa: F401


class Attribute(Locatable):
    """
        The attribute base class for entity attributes.

        @param entity: The entity this attribute belongs to
    """

    def __init__(self, entity: "Entity", value_type: "Type", name: str, multi: bool=False, nullable: bool=False) -> None:
        Locatable.__init__(self)
        self.__name: str = name
        entity.add_attribute(self)
        self.__entity = entity
        self.__type = value_type
        self.__multi = multi
        self.__nullallble = nullable
        self.low = 0 if nullable else 1
        self.comment = None  # type: Optional[str]
        self.end: Optional[RelationAttribute] = None

    def get_type(self) -> "Type":
        """
            Get the type of this data item
        """
        return self.__type

    type: "Type" = property(get_type)

    def get_name(self) -> str:
        """
            Get the name of the attribute. This is the name this attribute
            is associated with in the entity.
        """
        return self.__name

    name = property(get_name)

    def __hash__(self) -> "int":
        """
            The hash of this object is based on the name of the attribute
        """
        return hash(self.__name)

    def __repr__(self) -> str:
        return self.__name

    def get_entity(self) -> "Entity":
        """
            Return the entity this attribute belongs to
        """
        return self.__entity

    entity = property(get_entity)

    def validate(self, value: object) -> None:
        """
            Validate a value that is going to be assigned to this attribute
        """
        if not isinstance(value, Unknown):
            self.type.validate(value)

    def get_new_result_variable(self, instance: "Instance", queue: QueueScheduler) -> ResultVariable:
        if self.__multi:
            mytype = (TypedList(self.__type))
        else:
            mytype = (self.__type)

        out: ResultVariable["Instance"]

        if(self.__nullallble):
            # be a 0-1 relation
            self.end = None
            self.low = 0
            self.high = 1
            out = OptionVariable(self, instance, queue)
            mytype = NullableType(mytype)
        else:
            out = ResultVariable()

        out.set_type(mytype)
        return out

    def is_optional(self) -> bool:
        return self.__nullallble

    def is_multi(self) -> bool:
        return self.__multi

    def final(self, excns: List[Exception]) -> None:
        pass


class RelationAttribute(Attribute):
    """
        An attribute that is a relation
    """
    def __init__(self, entity: "Entity", value_type: "Type", name: str) -> None:
        Attribute.__init__(self, entity, value_type, name)
        self.end: Optional[RelationAttribute] = None
        self.low = 1
        self.high = 1
        self.depends = False
        self.source_annotations = []
        self.target_annotations = []

    def __str__(self) -> str:
        return "%s.%s" % (self.get_entity().get_full_name(), self.name)

    def __repr__(self) -> str:
        return "[%d:%s] %s" % (self.low, self.high if self.high is not None else "", self.name)

    def set_multiplicity(self, values: "Tuple[int, int]") -> None:
        """
            Set the multiplicity of this end
        """
        self.low = values[0]
        self.high = values[1]

    def get_new_result_variable(self, instance: "Instance", queue: QueueScheduler) -> ResultVariable:
        if self.low == 1 and self.high == 1:
            out = AttributeVariable(self, instance)  # type: ResultVariable
        elif self.low == 0 and self.high == 1:
            out = OptionVariable(self, instance, queue)  # type: ResultVariable
        else:
            out = ListVariable(self, instance, queue)  # type: ResultVariable
        out.set_type(self.get_type())
        return out

    def is_optional(self) -> bool:
        return self.low == 0

    def is_multi(self) -> bool:
        return self.high != 1

    def final(self, excns: List[Exception]) -> None:
        for rv in self.source_annotations:
            try:
                if isinstance(rv.get_value(), Unknown):
                    excns.append(TypingException(self, "Relation annotation can not be Unknown"))
            except RuntimeException as e:
                excns.append(e)
        for rv in self.target_annotations:
            try:
                if isinstance(rv.get_value(), Unknown):
                    excns.append(TypingException(self, "Relation annotation can not be Unknown"))
            except RuntimeException as e:
                excns.append(e)



"""
    Copyright 2016 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import logging
import os
import re
import urllib

from inmanta import data
from inmanta.agent.handler import provider, ResourceHandler, HandlerContext, CRUDHandler, ResourcePurged
from inmanta.execute.util import Unknown
from inmanta.resources import Resource, PurgeableResource, resource, ResourceNotFoundExcpetion, IgnoreResourceException


LOGGER = logging.getLogger(__name__)
FILE_SOURCE = "imp-module-source:"


def generate_content(content_list, seperator):
    """
        Generate a sorted list of content to prefix or suffix a file
    """
    sort_list = []
    for content in content_list:
        if content.sorting_key is None:
            sort_list.append((content.value, content.value))
        else:
            sort_list.append((content.sorting_key, content.value))

    sort_list.sort(key=lambda tup: tup[0])
    return seperator.join([x[1] for x in sort_list]) + seperator


def store_file(exporter, obj):
    content = obj.content
    if isinstance(content, Unknown):
        return content

    elif content.startswith(FILE_SOURCE):
        parts = urllib.parse.urlparse(content[len(FILE_SOURCE):])

        if parts.scheme == "file":
            with open(parts.path, "rb") as fd:
                content = fd.read()

        elif parts.scheme == "tmp":
            with open(parts.path, "rb") as fd:
                content = fd.read()

        else:
            raise Exception("%s scheme not support for file %s" %
                            (parts.scheme, parts.path))

    if len(obj.prefix_content) > 0:
        content = generate_content(obj.prefix_content, obj.content_seperator) + obj.content_seperator + content
    if len(obj.suffix_content) > 0:
        content += obj.content_seperator + generate_content(obj.suffix_content, obj.content_seperator)

    return exporter.upload_file(content)


@resource("std::Service", agent="host.name", id_attribute="name")
class Service(Resource):
    """
        This class represents a service on a system.
    """
    fields = ("onboot", "state", "name", "reload")


@resource("std::File", agent="host.name", id_attribute="path")
class File(PurgeableResource):
    """
        A file on a filesystem
    """
    fields = ("path", "owner", "hash", "group", "permissions", "reload")
    map = {"hash": store_file, "permissions": lambda y, x: int(x.mode)}


@resource("std::Directory", agent="host.name", id_attribute="path")
class Directory(PurgeableResource):
    """
        A directory on a filesystem
    """
    fields = ("path", "owner", "group", "permissions", "reload")
    map = {"permissions": lambda y, x: int(x.mode)}


@resource("std::Package", agent="host.name", id_attribute="name")
class Package(Resource):
    """
        A software package installed on an operating system.
    """
    fields = ("name", "state", "reload")


@resource("std::Symlink", agent="host.name", id_attribute="target")
class Symlink(PurgeableResource):
    """
        A symbolic link on the filesystem
    """
    fields = ("source", "target", "reload")


@resource("std::AgentConfig", agent="agent", id_attribute="agentname")
class AgentConfig(PurgeableResource):
    """
        A resource that can modify the agentmap for autostarted agents
    """
    fields = ("agentname", "uri", "autostart")

    @staticmethod
    def get_autostart(exp, obj):
        try:
            if not obj.autostart:
                raise IgnoreResourceException()
        except Exception as e:
            # When this attribute is not set, also ignore it
            raise IgnoreResourceException()
        return obj.autostart


@provider("std::File", name="posix_file")
class PosixFileProvider(ResourceHandler):
    """
        This handler can deploy files on a unix system
    """
    def check_resource(self, ctx: HandlerContext, resource: File) -> File:
        current = resource.clone(purged=False, reload=resource.reload, hash=0)

        if not self._io.file_exists(resource.path):
            current.purged = True
            current.hash = 0
            current.owner = None
            current.group = None
            current.permissions = None

        else:
            current.hash = self._io.hash_file(resource.path)

            # upload the previous version for back up and for generating a diff!
            content = self._io.read_binary(resource.path)

            if not self.stat_file(current.hash):
                self.upload_file(current.hash, content)

            for key, value in self._io.file_stat(resource.path).items():
                setattr(current, key, value)

        return current

    def do_changes(self, ctx: HandlerContext, resource: File, changes: dict) -> None:
        if "purged" in changes and changes["purged"]["desired"]:
            self._io.remove(resource.path)
            ctx.set_purged()
            return

        created = False
        updated = False
        if "hash" in changes:
            # write the new version
            dir_name = os.path.dirname(resource.path)
            if self._io.file_exists(dir_name):
                data = self.get_file(resource.hash)
                self._io.put(resource.path, data)
            else:
                raise Exception("Parent of %s does not exist, unable to write file." % resource.path)

            if "purged" in changes:
                created = True
            else:
                updated = True

        if "permissions" in changes:
            if not self._io.file_exists(resource.path):
                raise Exception("Cannot change permissions of %s because does not exist" % resource.path)

            self._io.chmod(resource.path, str(resource.permissions))
            updated = True

        if "owner" in changes or "group" in changes:
            if not self._io.file_exists(resource.path):
                raise Exception("Cannot change ownership of %s because does not exist" % resource.path)

            self._io.chown(resource.path, resource.owner, resource.group)
            updated = True

        if created:
            ctx.set_created()
        elif updated:
            ctx.set_updated()

    def snapshot(self, resource):
        return self._io.read_binary(resource.path)


@provider("std::Service", name="systemd")
class SystemdService(ResourceHandler):
    """
        A handler for services on systems that use systemd
    """
    def __init__(self, agent, io=None):
        super().__init__(agent, io)

        self._systemd_path = None

    def available(self, resource):
        if self._io.file_exists("/usr/bin/systemctl"):
            self._systemd_path = "/usr/bin/systemctl"

        elif self._io.file_exists("/bin/systemctl"):
            self._systemd_path = "/bin/systemctl"

        return self._systemd_path is not None

    def check_resource(self, ctx, resource):
        current = resource.clone()

        exists = self._io.run(self._systemd_path, ["status", "%s.service" % resource.name])[0]

        if re.search('Loaded: error', exists):
            raise ResourceNotFoundExcpetion("The %s service does not exist" % resource.name)

        running = self._io.run(self._systemd_path, ["is-active", "%s.service" % resource.name])[2] == 0
        enabled = self._io.run(self._systemd_path, ["is-enabled", "%s.service" % resource.name])[2] == 0

        if running:
            current.state = "running"
        else:
            current.state = "stopped"

        current.onboot = enabled
        return current

    def can_reload(self):
        """
            Can this handler reload?
        """
        return True

    def do_reload(self, ctx, resource):
        """
            Reload this resource
        """
        ctx.info("Reloading service with reload-or-restart")
        self._io.run(self._systemd_path, ["reload-or-restart", "%s.service" % resource.name])

    def do_changes(self, ctx, resource, changes):
        if "state" in changes:
            action = "start"
            if changes["state"]["desired"] == "stopped":
                action = "stop"

            # start or stop the service
            result = self._io.run(self._systemd_path, [action, "%s.service" % resource.name])

            if re.search("^Failed", result[1]):
                raise Exception("Unable to %s %s: %s" % (action, resource.name, result[1]))

            ctx.set_updated()

        if "onboot" in changes:
            action = "enable"

            if not changes["onboot"]["desired"]:
                action = "disable"

            result = self._io.run(self._systemd_path, [action, "%s.service" % resource.name])
            ctx.set_updated()

            if re.search("^Failed", result[1]):
                raise Exception("Unable to %s %s: %s" % (action, resource.name, result[1]))


@provider("std::Service", name="redhat_service")
class ServiceService(ResourceHandler):
    """
        A handler for services on systems that use service
    """
    def available(self, resource):
        return (self._io.file_exists("/sbin/chkconfig") and self._io.file_exists("/sbin/service") and
                not self._io.file_exists("/usr/bin/systemctl"))

    def check_resource(self, ctx, resource):
        current = resource.clone()
        exists = self._io.run("/sbin/chkconfig", ["--list", resource.name])[0]

        if re.search('error reading information on service', exists):
            raise ResourceNotFoundExcpetion("The %s service does not exist" % resource.name)

        enabled = ":on" in self._io.run("/sbin/chkconfig", ["--list", resource.name])[0]
        running = self._io.run("/sbin/service", [resource.name, "status"])[2] == 0

        current.enabled = enabled
        if running:
            current.state = "running"
        else:
            current.state = "stopped"

        return current

    def can_reload(self):
        """
            Can this handler reload?
        """
        return True

    def do_reload(self, ctx, resource):
        """
            Reload this resource
        """
        self._io.run("/sbin/service", [resource.name, "reload"])

    def do_changes(self, ctx, resource, changes):
        if "state" in changes:
            action = "start"
            if changes["state"]["desired"] == "stopped":
                action = "stop"

            # start or stop the service
            result = self._io.run("/sbin/service", [resource.name, action])

            if re.search("^Failed", result[1]):
                raise Exception("Unable to %s %s: %s" % (action, resource.name, result[1]))

            ctx.set_updated()

        if "enabled" in changes:
            action = "on"

            if not changes["enabled"]["desired"]:
                action = "off"

            result = self._io.run("/sbin/chkconfig", [resource.name, action])
            ctx.set_updated()

            if re.search("^Failed", result[1]):
                raise Exception("Unable to %s %s: %s" % (action, resource.name, result[1]))


@provider("std::Package", name="yum")
class YumPackage(ResourceHandler):
    """
        A Package handler that uses yum
    """
    def available(self, resource):
        return (self._io.file_exists("/usr/bin/rpm") or self._io.file_exists("/bin/rpm")) \
            and (self._io.file_exists("/usr/bin/yum") or self._io.file_exists("/usr/bin/dnf"))

    def _parse_fields(self, lines):
        props = {}
        key = ""
        old_key = None
        for line in lines:
            if line.strip() == "":
                continue

            if line.strip() == "Available Packages":
                break

            result = re.search("^(.+) :\s+(.+)", line)
            if result is None:
                continue

            key, value = result.groups()
            key = key.strip()

            if key == "":
                props[old_key] += " " + value
            else:
                props[key] = value
                old_key = key

        return props

    def _run_yum(self, args):
        # todo: cache value
        if self._io.file_exists("/usr/bin/dnf"):
            return self._io.run("/usr/bin/dnf", ["-d", "0", "-e", "0", "-y"] + args)
        else:
            return self._io.run("/usr/bin/yum", ["-d", "0", "-e", "0", "-y"] + args)

    def check_resource(self, ctx, resource):
        yum_output = self._run_yum(["info", resource.name])
        lines = yum_output[0].split("\n")

        output = self._parse_fields(lines[1:])

        if "Repo" not in output:
            return {"state": "removed"}

        state = "removed"

        if output["Repo"] == "installed" or output["Repo"] == "@System":
            state = "installed"

        # check if there is an update
        yum_output = self._run_yum(["check-update", resource.name])
        lines = yum_output[0].split("\n")

        data = {"state": state, "version": output["Version"],
                "release": output["Release"], "update": None}

        if len(lines) > 0:
            parts = re.search("([^\s]+)\s+([^\s]+)\s+([^\s]+)", lines[0])
            if parts is not None and not lines[0].startswith("Security:"):
                version_str = parts.groups()[1]
                version, release = version_str.split("-")

                data["update"] = (version, release)

        return data

    def list_changes(self, ctx, resource):
        state = self.check_resource(ctx, resource)

        changes = {}
        if resource.state == "removed":
            if state["state"] != "removed":
                changes["state"] = (state["state"], resource.state)

        elif resource.state == "installed" or resource.state == "latest":
            if state["state"] != "installed":
                changes["state"] = (state["state"], "installed")

        if "update" in state and state["update"] is not None and resource.state == "latest":
            changes["version"] = ((state["version"], state["release"]), state["update"])

        return changes

    def _result(self, output):
        stdout = output[0].strip()
        error_msg = output[1].strip()
        if output[2] != 0:
            raise Exception("Yum failed: stdout:" + stdout + " errout: " + error_msg)

    def do_changes(self, ctx, resource, changes):
        if "state" in changes:
            if changes["state"][1] == "removed":
                self._result(self._run_yum(["remove", resource.name]))
                ctx.set_purged()

            elif changes["state"][1] == "installed":
                self._result(self._run_yum(["install", resource.name]))
                ctx.set_created()

        if "version" in changes:
            self._result(self._run_yum(["update", resource.name]))
            ctx.set_updated()


@provider("std::Directory", name="posix_directory")
class DirectoryHandler(ResourceHandler):
    """
        A handler for creating directories

        TODO: add recursive operations
    """
    def check_resource(self, ctx, resource):
        current = resource.clone(purged=False)

        if not self._io.file_exists(resource.path):
            current.purged = True

        else:
            for key, value in self._io.file_stat(resource.path).items():
                setattr(current, key, value)

        return current

    def do_changes(self, ctx, resource, changes):
        created = False
        updated = False
        if "purged" in changes:
            if changes["purged"]["desired"]:
                self._io.rmdir(resource.path)
                ctx.set_purged()
                return
            else:
                self._io.mkdir(resource.path)
                created = True

        if "permissions" in changes or created:
            mode = str(resource.permissions)
            self._io.chmod(resource.path, mode)
            updated = True

        if "owner" in changes or "group" in changes or created:
            self._io.chown(resource.path, resource.owner, resource.group)
            updated = True

        if created:
            ctx.set_created()
        elif updated:
            ctx.set_updated()


@provider("std::Symlink", name="posix_symlink")
class SymlinkProvider(ResourceHandler):
    """
        This handler can deploy symlinks on unix systems
    """
    def available(self, resource):
        return self._io.file_exists("/usr/bin/ln") or self._io.file_exists("/bin/ln")

    def check_resource(self, ctx, resource):
        current = resource.clone(purged=False)

        if not self._io.file_exists(resource.target):
            current.purged = True
            current.source = None
            current.target = None

        elif not self._io.is_symlink(resource.target):
            raise Exception("The target of resource %s already exists but is not a symlink." % resource)

        else:
            current.source = self._io.readlink(resource.target)

        return current

    def do_changes(self, ctx, resource, changes):
        if "purged" in changes:
            if changes["purged"]["desired"]:
                self._io.remove(resource.target)
                ctx.set_purged()

            else:
                self._io.symlink(resource.source, resource.target)
                ctx.set_created()

        elif "source" in changes:
            self._io.remove(resource.target)
            self._io.symlink(resource.source, resource.target)
            ctx.set_updated()


@provider("std::AgentConfig", name="agentrest")
class AgentConfigHandler(CRUDHandler):
    def _get_map(self) -> dict:
        def call():
            return self.get_client().get_setting(tid=self._agent.environment, id=data.AUTOSTART_AGENT_MAP)

        value = self.run_sync(call)
        return value.result["value"]

    def _set_map(self, agent_config: dict) -> None:
        def call():
            return self.get_client().set_setting(tid=self._agent.environment, id=data.AUTOSTART_AGENT_MAP, value=agent_config)

        return self.run_sync(call)

    def read_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
        agent_config = self._get_map()
        ctx.set("map", agent_config)

        if resource.agentname not in agent_config:
            raise ResourcePurged()

        resource.uri = agent_config[resource.agentname]

    def create_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
        agent_config = ctx.get("map")
        agent_config[resource.agentname] = resource.uri
        self._set_map(agent_config)

    def delete_resource(self, ctx: HandlerContext, resource: AgentConfig) -> None:
        agent_config = ctx.get("map")
        del agent_config[resource.agentname]
        self._set_map(agent_config)

    def update_resource(self, ctx: HandlerContext, changes: dict, resource: AgentConfig) -> None:
        agent_config = ctx.get("map")
        agent_config[resource.agentname] = resource.uri
        self._set_map(agent_config)



"""
    Copyright 2017-2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

import os
import sys
import logging
import glob
import imp

from inmanta import const
from inmanta.execute import scheduler
from inmanta.ast import Namespace, LocatableString, Range
from inmanta.ast.statements.define import DefineEntity, DefineRelation, PluginStatement
from inmanta.module import Project
from inmanta.plugins import PluginMeta

LOGGER = logging.getLogger(__name__)


def do_compile(refs={}):
    """
        Run run run
    """
    project = Project.get()
    compiler = Compiler(os.path.join(project.project_path, project.main_file), refs=refs)

    LOGGER.debug("Starting compile")

    (statements, blocks) = compiler.compile()
    sched = scheduler.Scheduler()
    success = sched.run(compiler, statements, blocks)

    LOGGER.debug("Compile done")

    if not success:
        sys.stderr.write("Unable to execute all statements.\n")
    return (sched.get_types(), compiler.get_ns())


def anchormap(refs={}):
    """
        Run run run
    """
    project = Project.get()
    compiler = Compiler(os.path.join(project.project_path, project.main_file), refs=refs)

    LOGGER.debug("Starting compile")

    (statements, blocks) = compiler.compile()
    sched = scheduler.Scheduler()
    return sched.anchormap(compiler, statements, blocks)


class Compiler(object):
    """
        An inmanta compiler

        @param options: Options passed to the application
        @param config: The parsed configuration file
    """

    def __init__(self, cf_file="main.cf", refs={}):
        self.__init_cf = "_init.cf"

        self.__cf_file = cf_file
        self.__root_ns = None
        self.refs = refs

    def get_plugins(self):
        return self.plugins

    def is_loaded(self):
        """
            Is everything loaded and the namespace structure built?
        """
        return self.__root_ns is not None

    def get_ns(self):
        """
            Get the root namespace
        """
        if not self.is_loaded():
            self.load()
        return self.__root_ns

    ns = property(get_ns)

    def read(self, path):
        """
            Return the content of the given file
        """
        with open(path, "r") as file_d:
            return file_d.read()

    def _load_plugins(self, plugin_dir, namespace):
        """
            Load all modules in plugin_dir
        """
        if not os.path.exists(os.path.join(plugin_dir, "__init__.py")):
            raise Exception("The plugin directory %s should be a valid python package with a __init__.py file" % plugin_dir)

        mod_name = ".".join(namespace.to_path())
        imp.load_package(mod_name, plugin_dir)

        for py_file in glob.glob(os.path.join(plugin_dir, "*.py")):
            if not py_file.endswith("__init__.py"):
                # name of the python module
                sub_mod = mod_name + "." + os.path.basename(py_file).split(".")[0]

                # create a namespace for the submodule
                new_ns = Namespace(sub_mod.split(".")[-1])
                new_ns.parent = namespace
                self.graph.add_namespace(new_ns, namespace)

                # load the python file
                imp.load_source(sub_mod, py_file)

    def compile(self):
        """
            This method will compile and prepare everything to start evaluation
            the configuration specification.

            This method will:
            - load all namespaces
            - compile the __config__ namespace
            - start resolving it and importing unknown namespaces
        """
        project = Project.get()
        self.__root_ns = project.get_root_namespace()

        project.load()
        statements, blocks = project.get_complete_ast()

        # load plugins
        for name, cls in PluginMeta.get_functions().items():

            mod_ns = cls.__module__.split(".")
            if mod_ns[0] != const.PLUGINS_PACKAGE:
                raise Exception("All plugin modules should be loaded in the impera_plugins package not in %s" % cls.__module__)

            mod_ns = mod_ns[1:]

            ns = self.__root_ns
            for part in mod_ns:
                if ns is None:
                    break
                ns = ns.get_child(part)

            if ns is None:
                raise Exception("Unable to find namespace for plugin module %s" % (cls.__module__))

            cls.namespace = ns

            name = name.split("::")[-1]
            statement = PluginStatement(ns, name, cls)
            statements.append(statement)

        # add the entity type (hack?)
        ns = self.__root_ns.get_child_or_create("std")
        nullrange = Range("internal", 1, 0, 0, 0)
        entity = DefineEntity(ns, LocatableString("Entity", nullrange, 0, ns),
                              "The entity all other entities inherit from.", [], [])
        str_std_entity = LocatableString("std::Entity", nullrange, 0, ns)

        requires_rel = DefineRelation((str_std_entity, LocatableString("requires", nullrange, 0, ns), [0, None], False),
                                      (str_std_entity, LocatableString("provides", nullrange, 0, ns), [0, None], False))
        requires_rel.namespace = self.__root_ns.get_ns_from_string("std")

        statements.append(entity)
        statements.append(requires_rel)
        return (statements, blocks)



"""
    Copyright 2017 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""

# pylint: disable-msg=W0613

from . import ReferenceStatement
from inmanta.ast.type import List, Dict
from inmanta.ast.statements import AssignStatement, ExpressionStatement, Statement, Resumer
from inmanta.execute.runtime import ExecutionUnit, ResultVariable, HangUnit, Instance, Resolver, QueueScheduler
from inmanta.execute.util import Unknown
from inmanta.ast import RuntimeException, AttributeException, DuplicateException, TypingException, LocatableString,\
    TypeReferenceAnchor, KeyException
from inmanta.ast.attribute import RelationAttribute
import typing

try:
    from typing import TYPE_CHECKING
except ImportError:
    TYPE_CHECKING = False


if TYPE_CHECKING:
    from inmanta.ast.variables import Reference  # noqa: F401


class CreateList(ReferenceStatement):
    """
        Create list of values
    """

    def __init__(self, items: typing.List[ExpressionStatement]) -> None:
        ReferenceStatement.__init__(self, items)
        self.items = items

    def execute(self,
                requires: typing.Dict[object, object],
                resolver: Resolver,
                queue: QueueScheduler) -> object:
        """
            Create this list
        """
        qlist = List()

        for i in range(len(self.items)):
            value = self.items[i]
            qlist.append(value.execute(requires, resolver, queue))

        return qlist

    def execute_direct(self, requires):
        qlist = List()

        for i in range(len(self.items)):
            value = self.items[i]
            qlist.append(value.execute_direct(requires))

        return qlist

    def __repr__(self) -> str:
        return "List()"


class CreateDict(ReferenceStatement):

    def __init__(self, items: typing.List[typing.Tuple[str, ReferenceStatement]]) -> None:
        ReferenceStatement.__init__(self, [x[1] for x in items])
        self.items = items
        seen = {}  # type: typing.Dict[str,ReferenceStatement]
        for x, v in items:
            if x in seen:
                raise DuplicateException(v, seen[x], "duplicate key in dict %s" % x)
            seen[x] = v

    def execute(self, requires: typing.Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        """
            Create this list
        """
        qlist = Dict()

        for i in range(len(self.items)):
            key, value = self.items[i]
            qlist[key] = value.execute(requires, resolver, queue)

        return qlist

    def __repr__(self) -> str:
        return "Dict()"


class SetAttribute(AssignStatement, Resumer):
    """
        Set an attribute of a given instance to a given value
    """

    def __init__(self, instance: "Reference", attribute_name: str, value: ExpressionStatement, list_only: bool=False) -> None:
        AssignStatement.__init__(self, instance, value)
        self.instance = instance
        self.attribute_name = attribute_name
        self.value = value
        self.list_only = list_only

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        reqs = self.instance.requires_emit(resolver, queue)
        HangUnit(queue, resolver, reqs, ResultVariable(), self)

    def resume(self,
               requires: typing.Dict[object, object],
               resolver: Resolver,
               queue: QueueScheduler,
               target: ResultVariable) -> None:
        instance = self.instance.execute(requires, resolver, queue)
        if not isinstance(instance, Instance):
            raise TypingException(self, "The object at %s is not an Entity but a %s with value %s" %
                                  (self.instance, type(instance), instance))
        var = instance.get_attribute(self.attribute_name)
        if self.list_only and not var.is_multi():
            raise TypingException(self, "Can not use += on relations with multiplicity 1")
        reqs = self.value.requires_emit_gradual(resolver, queue, var)
        SetAttributeHelper(queue, resolver, var, reqs, self.value, self, instance, self.attribute_name)

    def __str__(self) -> str:
        return "%s.%s = %s" % (str(self.instance), self.attribute_name, str(self.value))


class SetAttributeHelper(ExecutionUnit):

    def __init__(self,
                 queue_scheduler: QueueScheduler,
                 resolver: Resolver,
                 result: ResultVariable,
                 requires: typing.Dict[object, ResultVariable],
                 expression: ExpressionStatement,
                 stmt: Statement,
                 instance: Instance,
                 attribute_name: str) -> None:
        ExecutionUnit.__init__(self, queue_scheduler, resolver, result, requires, expression)
        self.stmt = stmt
        self.instance = instance
        self.attribute_name = attribute_name

    def execute(self) -> None:
        try:
            ExecutionUnit._unsafe_execute(self)
        except RuntimeException as e:
            e.set_statement(self.stmt, False)
            raise AttributeException(self.stmt, self.instance, self.attribute_name, e)

    def __str__(self) -> str:
        return str(self.stmt)


class Assign(AssignStatement):
    """
        This class represents the assignment of a value to a variable -> alias

        @param name: The name of the value
        @param value: The value that is to be assigned to the variable

        uses:          value
        provides:      variable
    """

    def __init__(self, name: str, value: ExpressionStatement) -> None:
        AssignStatement.__init__(self, None, value)
        self.name = name
        self.value = value

    def requires(self) -> typing.List[str]:
        return self.value.requires()

    def emit(self, resolver: Resolver, queue: QueueScheduler) -> None:
        target = resolver.lookup(self.name)
        assert isinstance(target, ResultVariable)
        reqs = self.value.requires_emit(resolver, queue)
        ExecutionUnit(queue, resolver, target, reqs, self.value)

    def __repr__(self) -> str:
        return "Assign(%s, %s)" % (self.name, self.value)


class MapLookup(ReferenceStatement):
    """
        Lookup a value in a dict
    """

    def __init__(self,
                 themap: ExpressionStatement,
                 key: ExpressionStatement
                 ):
        super(MapLookup, self).__init__([themap, key])
        self.themap = themap
        self.key = key
        self.location = themap.get_location().merge(key.location)

    def execute(self, requires: typing.Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        mapv = self.themap.execute(requires, resolver, queue)
        if not isinstance(mapv, dict):
            raise TypingException(self, "dict lookup is only possible on dicts, %s is not an object" % mapv)

        keyv = self.key.execute(requires, resolver, queue)
        if not isinstance(keyv, str):
            raise TypingException(self, "dict keys must be string, %s is not a string" % keyv)

        if keyv not in mapv:
            raise KeyException(self, "key %s not found in dict, options are [%s]" % (keyv, ",".join(mapv.keys())))

        return mapv[keyv]

    def __repr__(self) -> str:
        return "%s[%s]" % (repr(self.themap), repr(self.key))


class IndexLookup(ReferenceStatement, Resumer):
    """
        Lookup a value in a dictionary
    """

    def __init__(self,
                 index_type: LocatableString,
                 query: typing.List[typing.Tuple[LocatableString, ExpressionStatement]]) -> None:
        ReferenceStatement.__init__(self, [v for (_, v) in query])
        self.index_type = str(index_type)
        self.anchors.append(TypeReferenceAnchor(index_type.get_location(), index_type.namespace, str(index_type)))
        self.query = [(str(n), e) for n, e in query]

    def normalize(self) -> None:
        ReferenceStatement.normalize(self)
        self.type = self.namespace.get_type(self.index_type)

    def requires_emit(self, resolver: Resolver, queue: QueueScheduler) -> typing.Dict[object, ResultVariable]:
        sub = ReferenceStatement.requires_emit(self, resolver, queue)
        temp = ResultVariable()
        temp.set_type(self.type)
        temp.set_provider(self)
        HangUnit(queue, resolver, sub, temp, self)
        return {self: temp}

    def resume(self,
               requires: typing.Dict[object, object],
               resolver: Resolver,
               queue: QueueScheduler,
               target: ResultVariable) -> None:
        self.type.lookup_index([(k, v.execute(requires, resolver, queue)) for (k, v) in self.query], self, target)

    def execute(self, requires: typing.Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        return requires[self]

    def __repr__(self) -> str:
        """
            The representation of this statement
        """
        return "%s[%s]" % (self.index_type, self.query)


class ShortIndexLookup(IndexLookup):
    """lookup of the form
vm = ip::Host(...)
file = std::File(host=vm, path="/etc/motd", ...)

vm.files[path="/etc/motd"]
    """

    def __init__(self, rootobject: ExpressionStatement,
                 relation: LocatableString, query:
                 typing.List[typing.Tuple[LocatableString, ExpressionStatement]]):
        ReferenceStatement.__init__(self, [v for (_, v) in query] + [rootobject])
        self.rootobject = rootobject
        self.relation = str(relation)
        self.querypart: typing.List[typing.Tuple[str, ExpressionStatement]] = [(str(n), e) for n, e in query]

    def normalize(self) -> None:
        ReferenceStatement.normalize(self)
        # currently there is no way to get the type of an expression prior to evaluation
        self.type = None

    def resume(self,
               requires: typing.Dict[object, object],
               resolver: Resolver,
               queue: QueueScheduler,
               target: ResultVariable) -> None:
        root_object = self.rootobject.execute(requires, resolver, queue)

        if not isinstance(root_object, Instance):
            raise TypingException(self, "short index lookup is only possible one objects, %s is not an object" % root_object)

        from_entity = root_object.get_type()

        relation = from_entity.get_attribute(self.relation)
        if not isinstance(relation, RelationAttribute):
            raise TypingException(self, "short index lookup is only possible on relations, %s is an attribute" % relation)

        if relation.end is None:
            raise TypingException(
                self,
                "short index lookup is only possible on bi-drectional relations, %s is unidirectional" % relation)

        self.type = relation.get_type()

        args: typing.List[typing.Tuple[str, object]] = [(relation.end.name, root_object)]
        args = args + [(k, v.execute(requires, resolver, queue)) for (k, v) in self.querypart]

        self.type.lookup_index(args, self, target)

    def __repr__(self) -> str:
        """
            The representation of this statement
        """
        return "%s.%s[%s]" % (self.rootobject, self.relation, self.querypart)


class StringFormat(ReferenceStatement):
    """
        Create a new string by doing a string interpolation
    """

    def __init__(self, format_string: str, variables: typing.List[typing.Tuple["Reference", str]]) -> None:
        ReferenceStatement.__init__(self, [k for (k, _) in variables])
        self._format_string = format_string
        self._variables = variables

    def execute(self, requires: typing.Dict[object, object], resolver: Resolver, queue: QueueScheduler) -> object:
        result_string = self._format_string
        for _var, str_id in self._variables:
            value = _var.execute(requires, resolver, queue)
            if isinstance(value, Unknown):
                return Unknown(self)
            if isinstance(value, float) and (value - int(value)) == 0:
                value = int(value)

            result_string = result_string.replace(str_id, str(value))

        return result_string

    def __repr__(self) -> str:
        return "Format(%s)" % self._format_string
